#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
**中文文档**

object file io是一个将Python对象对单个本地文件的I/O



"""

import os
import time
import zlib
import base64
import shutil
import logging
import inspect

# logging util
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
logger.addHandler(stream_handler)


def prt_console(message, verbose):
    """Print message to console, if ``verbose`` is True. 
    """
    if verbose:
        logger.info(message)


# dump, safe_dump, load
def _dump(obj, abspath,
          dumper_func=None,
          compress=True,
          overwrite=False,
          verbose=False,
          **kwargs):
    """Dump object to file.

    :param abspath: The file path you want dump to.
    :type abspath: str

    :param dumper_func: A dumper function that takes an object as input, return
        binary.
    :type dumper_func: callable function

    :param compress: default ``False``. If True, then compress binary.
    :type compress: bool

    :param overwrite: default ``False``, If ``True``, when you dump to 
      existing file, it silently overwrite it. If ``False``, an alert 
      message is shown. Default setting ``False`` is to prevent overwrite 
      file by mistake.
    :type overwrite: boolean

    :param verbose: default True, help-message-display trigger.
    :type verbose: boolean
    """
    if not inspect.isfunction(dumper_func):
        raise TypeError("dumper_func has to be a function take object as input "
                        "and return binary!")

    prt_console("\nDump to '%s' ..." % abspath, verbose)
    if os.path.exists(abspath):
        if not overwrite:
            prt_console(
                "    Stop! File exists and overwrite is not allowed",
                verbose,
            )
            return

    st = time.clock()

    b = dumper_func(obj, **kwargs)
    if compress:
        b = zlib.compress(b)
    with open(abspath, "wb") as f:
        f.write(b)

    elapsed = time.clock() - st
    prt_console("    Complete! Elapse %.6f sec." % elapsed, verbose)

    return b


def _safe_dump(obj, abspath,
               dumper_func=None,
               compress=True,
               verbose=False,
               **kwargs):
    """A stable version of :func:`._dump`, this method silently overwrite 
    existing file.

    There's a issue with :func:`_dump`: If your program is 
    interrupted while writing, you got an incomplete file, and you also 
    lose the original file. So this method write json to a temporary file 
    first, then rename to what you expect, and silently overwrite old one. 
    This way can guarantee atomic write operation.

    **中文文档**

    在对文件进行写入时, 如果程序中断, 则会留下一个不完整的文件。如果使用了
    覆盖式写入, 则我们即没有得到新文件, 同时也丢失了原文件。所以为了保证
    写操作的原子性(要么全部完成, 要么全部都不完成), 更好的方法是: 首先将
    文件写入一个临时文件中, 完成后再讲文件重命名, 覆盖旧文件。这样即使中途
    程序被中断, 也仅仅是留下了一个未完成的临时文件而已, 不会影响原文件。
    """
    abspath_temp = "%s.tmp" % abspath
    b = _dump(
        obj,
        abspath_temp,
        dumper_func=dumper_func,
        compress=compress,
        overwrite=True,
        verbose=verbose,
        **kwargs
    )
    shutil.move(abspath_temp, abspath)

    return b


def _load(abspath,
          loader_func=None,
          decompress=True,
          verbose=False,
          **kwargs):
    """load object from file.

    :param abspath: The file path you want load from.
    :type abspath: str

    :param loader_func: A loader function that takes binary as input, return
        an object.
    :type loader_func: callable function

    :param decompress: default ``False``. If True, then decompress binary.
    :type decompress: bool

    :param verbose: default True, help-message-display trigger.
    :type verbose: boolean
    """
    if not inspect.isfunction(loader_func):
        raise TypeError("loader_func has to be a function take binary as input "
                        "and return an object!")

    prt_console("\nLoad from '%s' ..." % abspath, verbose)
    if not os.path.exists(abspath):
        raise ValueError("'%s' doesn't exist." % abspath)

    st = time.clock()

    with open(abspath, "rb") as f:
        b = f.read()
        if decompress:
            b = zlib.decompress(b)
    obj = loader_func(b, **kwargs)

    elapsed = time.clock() - st
    prt_console("    Complete! Elapse %.6f sec." % elapsed, verbose)

    return obj


if __name__ == "__main__":
    import pickle
    import functools

    def pickle_dumper(obj):
        return pickle.dumps(obj)

    def pickle_loader(b):
        return pickle.loads(b)

    def test_dumper_loader():
        obj = dict(a=1, b=2)
        b = pickle_dumper(obj)
        obj = pickle_loader(b)

    test_dumper_loader()

    def test_partial():
        dump = functools.partial(_dump, dumper_func=pickle_dumper)
        safe_dump = functools.partial(_safe_dump, dumper_func=pickle_dumper)
        load = functools.partial(_load, loader_func=pickle_loader)

        obj = dict(a=1, b=2, c=3)
        b = dump(obj, "data.pk", verbose=True)
        b = safe_dump(obj, "data.pk", verbose=True)
        obj1 = load("data.pk", verbose=True)
        assert obj == obj1

    test_partial()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import types
from multiprocessing import cpu_count
from multiprocessing.dummy import Pool

try:
    from .task import Task
    from .per_class_logger import ClassWithLogger
    from .util.fingerprint import hash_data
    from .pkg.sfm.exception_mate import get_last_exc_info
except:  # pragma: no cover
    from pytq.task import Task
    from pytq.per_class_logger import ClassWithLogger
    from pytq.util.fingerprint import hash_data
    from pytq.pkg.sfm.exception_mate import get_last_exc_info


class BaseScheduler(ClassWithLogger):
    """
    All Scheduler has to inherit from this base class.
    """

    def __init__(self, logger=None):
        super(BaseScheduler, self).__init__(logger=logger)

        # link hash input method
        try:
            self.user_hash_input(None)
            self._hash_input = self.user_hash_input
        except NotImplementedError:
            self._hash_input = self._default_hash_input
        except:
            self._hash_input = self.user_hash_input

        # link duplicate test method
        try:
            self.user_is_duplicate(None)
            self._is_duplicate = self.user_is_duplicate
        except NotImplementedError:
            self._is_duplicate = self._default_is_duplicate
        except:
            self._is_duplicate = self.user_is_duplicate

        # link pre process method
        try:
            self.user_pre_process(None)
            self._pre_process = self.user_pre_process
        except NotImplementedError:
            self._pre_process = self._default_pre_process
        except:
            self._pre_process = self.user_pre_process

        # link post process method
        try:
            self.user_post_process(None)
            self._post_process = self.user_post_process
        except NotImplementedError:
            self._post_process = self._default_post_process
        except:
            self._post_process = self.user_post_process

    def _default_hash_input(self, input_data):
        """
        Default hash method to get a identical fingerprint for input data.

        By default its: pickle the data and md5 it.

        This method will be used when :meth:`BaseScheduler.user_hash_input`
        are not defined.

        :returns: fingerprint for ``input_data``
        :rtype: string.
        """
        return hash_data(input_data)

    def user_hash_input(self, input_data):
        """
        (Optional) Get identical fingerprint for input data.

        :returns: fingerprint for ``input_data``
        :rtype: string.
        """
        raise NotImplementedError

    def _hash_input(self, input_data):
        """
        The real hashing method will be called.

        :returns: fingerprint for ``input_data``
        :rtype: string.
        """
        raise NotImplementedError

    def user_is_duplicate(self, task):
        """
        (Optional) Check if a task is duplicate.

        :param task:
        :return: return True, when it's a duplicate item.
        :rtype: boolean.
        """
        raise NotImplementedError

    def _default_is_duplicate(self, task):
        """
        Default duplicate test method, always not duplicate.

        :param task:
        :return:
        """
        return False  # Alwasy not duplicate

    def _is_duplicate(self, task):
        """
        The real duplicate test method will be called.

        :returns: boolean. return True, when it's a duplicate item.
        """
        raise NotImplementedError

    def _remove_duplicate(self, input_data_queue):
        """
        Remove duplicate input_data.

        :param input_data_queue:
        :returns: task_queue.
        :rtype: types.GeneratorType.

        **中文文档**

        移除那些重复的输入数据。
        """
        is_generator = isinstance(
            input_data_queue, types.GeneratorType)

        if not is_generator:
            left_counter = len(input_data_queue)

        nth_counter = 0
        for input_data in input_data_queue:
            if is_generator:
                left_counter = None
            else:
                left_counter -= 1

            task = Task(
                id=self._hash_input(input_data),
                input_data=input_data,
                nth_counter=nth_counter,
                left_counter=left_counter,
            )

            if not self._is_duplicate(task):
                nth_counter += 1
                yield task

    _default_batch_pre_process = _remove_duplicate
    """
    A method will be called to pre process task queue before doing any real
    per task process. Usually it can be duplicate filter, statistic check.
    """

    def user_pre_process(self, task):
        """
        (Optional) Defines the action that before the
        :meth:`BaseScheduler.user_process() been called.

        :param task: :class:`pytq.task.Task` instance.
        """
        raise NotImplementedError

    def _default_pre_process(self, task):
        """
        Default behavior of user_pre_process, do nothing.
        """
        pass

    def _pre_process(self, task):
        """
        The real method will be called for pre_process.
        """
        raise NotImplementedError

    def user_post_process(self, task):
        """
        (Optional) Defines the action that after the
        :meth:`BaseScheduler.user_process() been called.

        :param task: :class:`pytq.task.Task` instance.
        """
        raise NotImplementedError

    def _default_post_process(self, task):
        """
        Default behavior of user_post_process, do nothing.
        """
        pass

    def _post_process(self, task):
        """
        The real method will be called for post_process.
        """
        raise NotImplementedError

    def user_process(self, input_data):
        """
        (Required) Defines the logic that process the input_data, returns
        output_data

        :param input_data:
        :return: the output_data
        """
        raise NotImplementedError

    def _process(self, task):
        """
        The real processing method will be called.

        **中文文档**

        处理数据。包含, 哈希, 处理, 标记完成三个步骤
        """
        self.info(task.progress_msg())

        try:
            if task.pre_process is None:
                self._pre_process(task)
            else:
                task._pre_process()

            output_data = self.user_process(task.input_data)
            task.output_data = output_data

            if task.post_process is None:
                self._post_process(task)
            else:
                task._post_process()

            self.info("Success!", 1)
        except Exception as e:
            self.info("Failed due to: %r" % get_last_exc_info(), 1)

    def _do_single_process(self, task_queue):
        """
        Execute single thread process.

        :param task_queue: task queue/list.
        """
        for task in task_queue:
            self._process(task)

    def _do_multi_process(self, task_queue):
        """
        Execute multi thread process.

        :param task_queue: task queue/list.
        """
        pool = Pool(processes=cpu_count())
        pool.map(self._process, task_queue)

    def do(self,
           input_data_queue,
           pre_process=None,
           multiprocess=False):
        """
        Do the real work.

        :param input_data_list: list of input data (or generator).
        :param quick_remove_duplicate: apply remove finished item (duplicate)
          filter  before we start the real work.
        :param multiprocess: trigger to use multiprocess.
        """
        if pre_process is None:
            task_queue = self._default_batch_pre_process(input_data_queue)
        else:
            task_queue = pre_process(input_data_queue)

        try:
            self.info("\nHas %s items todo." % len(input_data_queue))
        except:
            self.info("\nHas UNKNOWN items todo")

        if multiprocess:
            self._do_multi_process(task_queue)
        else:
            self._do_single_process(task_queue)

        self.info("Complete!")

    def clear_all(self):
        """
        Clear all data.

        **中文文档**

        重置Filter至初始状态。
        """
        raise NotImplementedError

    def __len__(self):
        raise NotImplementedError

    def __iter__(self):
        raise NotImplementedError

    def get(self, id):
        """
        Get output data by fingerprint of input_data.

        **中文文档**

        根据输入的指纹, 直接获得已经完成的输出数据。
        """
        raise NotImplementedError

    def get_output(self, input_data):
        """
        Get output data of the input_data.

        **中文文档**

        根据输入的数据, 直接获得已经完成的输出的数据。
        """
        return self.get(self._hash_input(input_data))

    def keys(self):
        return iter(self)

    def items(self):
        for key in self:
            yield (key, self.get(key))


class BaseDBTableBackedScheduler(BaseScheduler):
    """
    Scheduler that use database table as backend storage.

    - Task.id as primary_key
    - Other column / field for input_data, output_data storage.
    """

    def _get_finished_id_set(self):
        """
        (Required) A method that returns all saved id set.
        (For all processed input_data)

        :returns: a id set.
        :rtype: set.
        """
        raise NotImplementedError

    def _default_batch_pre_process(self, input_data_queue):
        finished_id_set = self._get_finished_id_set()

        is_generator = isinstance(input_data_queue, types.GeneratorType)
        # if its generator, then left_counter always be None
        if not is_generator:
            left_counter = len(input_data_queue)

        nth_counter = 0
        for input_data in input_data_queue:
            if is_generator:
                left_counter = None
            else:
                left_counter -= 1

            id = self._hash_input(input_data)
            if id not in finished_id_set:
                nth_counter += 1
                task = Task(
                    id=id,
                    input_data=input_data,
                    nth_counter=nth_counter,
                    left_counter=left_counter,
                )
                yield task



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pickle

try:
    from pytq.scheduler import Task, BaseDBTableBackedScheduler
except:  # pragma: no cover
    from pytq.scheduler import Task, BaseDBTableBackedScheduler


class MongoDBScheduler(BaseDBTableBackedScheduler):
    """

    :param collection: :class:`pymongo.Collection` instance.
    """
    collection = None

    def __init__(self, logger=None, collection=None):
        super(MongoDBScheduler, self).__init__(logger=logger)

        if collection is not None:
            self.collection = collection
        self._col = self.collection

        # link encode method
        try:
            self.user_encode(None)
            self._encode = self.user_encode
        except NotImplementedError:
            self._encode = self._default_encode
        except:
            self._encode = self.user_encode

        # link decode method
        try:
            self.user_decode(None)
            self._decode = self.user_decode
        except NotImplementedError:
            self._decode = self._default_decode
        except:
            self._decode = self.user_decode

    @property
    def collection(self):
        """
        Back-end mongodb collection.
        """
        raise NotImplementedError

    def _default_encode(self, obj):
        return pickle.dumps(obj)

    def user_encode(self, obj):
        """
        (Optional) User defined serializer for output_data.

        :returns: bytes or string.

        **中文文档**

        用于对处理结果序列化的函数。默认使用pickle。
        """
        raise NotImplementedError

    def _encode(self, obj):
        raise NotImplementedError

    def _default_decode(self, bytes_or_str):
        return pickle.loads(bytes_or_str)

    def user_decode(self, bytes_or_str):
        """
        (Optional) User defined deserializer for output_data.

        :returns: python object.

        **中文文档**

        用于对处理结果反序列化的函数。默认使用pickle。
        """
        raise NotImplementedError

    def _decode(self, bytes_or_str):
        raise NotImplementedError

    def _default_is_duplicate(self, task):
        """
        :param task:
        :return:
        """
        return self._col.find_one({"_id": task.key}) is not None

    def _get_finished_id_set(self):
        return set([
            doc["_id"] for doc in self._col.find({}, {"_id": True})
        ])

    def _default_post_process(self, task):
        """
        Save output_data into "out" field.
        """
        self._col.update(
            {"_id": task.id},
            {"$set": {"out": self._encode(task.output_data)}},
            upsert=True,
        )

    def __len__(self):
        return self._col.find().count()

    def __iter__(self):
        for doc in self._col.find():
            yield doc["_id"]

    def clear_all(self):
        self._col.remove({})

    def get_output(self, input_data):
        key = self.user_hash_input(input_data)
        return self._decode(self._col.find_one({"_id", key})["out"])

    def items(self):
        for doc in self._col.find():
            yield (doc["_id"], self._decode(doc["out"]))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import base64
import binascii


def encode_base64_urlsafe(text):
    """Convert any utf-8 string to url safe string using base64 encoding.

    **中文文档**

    将任意utf-8字符串用base64编码算法编码为纯数字和字母。
    """
    return base64.urlsafe_b64encode(text.encode("utf-8")).decode("utf-8")


def decode_base64_urlsafe(text):
    """Reverse operation of :func:`encode_base64_urlsafe`.

    **中文文档**

    将base64字符串解码为原字符串。
    """
    return base64.urlsafe_b64decode(text.encode("utf-8")).decode("utf-8")


def encode_hexstr(text):
    """Convert any utf-8 string to hex string.

    **中文文档**

    将任意utf-8字符串编码为16进制字符串。
    """
    return binascii.b2a_hex(text.encode("utf-8")).decode("utf-8")


def decode_hexstr(text):
    """Reverse operation of :func:`encode_hexstr`.

    **中文文档**

    将16进制字符串解码为原字符串。
    """
    return binascii.a2b_hex(text.encode("utf-8")).decode("utf-8")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

**中文文档**

mongocrawl是一个使用MongoDB作为数据持久层, 广度优先的定向爬虫框架。非常适合于
数据资源所在的Url是树状层级结构的爬虫项目。

举例来说, 我们希望抓取全国所有的邮政编码。而邮政编码数据网站的结构是:
'1个国家有a个省, 1个省有b个市, 1个市有c个区, 1个区有d个邮政编码', 类似这种的
层级结构。所以我们要遍历所有省的页面找到所有的市, 然后遍历所有的市的页面找到
所有的区... 这样层层向下遍历, 从而最终获得所有的邮政编码。
"""

import json
import pickle
import logging
import pymongo
import mongoengine
from crawlib.spider import spider
from loggerFactory import StreamOnlyLogger
from multiprocessing.dummy import Pool
from sfm.mongoengine_mate import ExtendedDocument

logging.getLogger("requests").setLevel(logging.WARNING)


class Task(ExtendedDocument):

    """

    **中文文档**

    由于是广度优先爬虫, 所以在对下一级的Url进行抓取时, 上一级全部的Url列表应该
    是已知的 (当然我们不一定需要完整的Url列表, 这一点之后再谈)。

    :meth:`Task.target_url`:

    每一个Url必然是不相同的。而我们有时可以不必使用Url作为主键, 因为有时Url可以
    通过运算得出, 而不必储存完整Url占据过多的磁盘资源。

    :attr:`Task.status`:

    status项标记了该资源是否被成功抓取过。初始状态下status = 0, 而再抓取过后可以
    将其修改为1。默认情况下mongocrawl只能识别0, 1两种状态。当然用户可以定义更多
    的状态码以应对不同的情况。

    对单个url资源进行抓取需要经过如下几个步骤:

    1. 获得url, :meth:`Task.target_url` 方法。用户需根据具体情况自行定义。
    2. 获得html, :meth:`Task.get_html` 方法。在网站无法通过restful的request请求
      获得html时需要重新自定义。
    3. 抓取data, :meth:`Task.fetch_data` 方法。用户需根据具体情况自行定义。
    4. 存入database, :meth:`Task.crawl_one` 方法。
    """
    _id = mongoengine.StringField(primary_key=True)
    status = mongoengine.IntField()
    json = mongoengine.StringField()
    pickle = mongoengine.BinaryField()

    meta = {"abstract": True}

    use_json = None

    logger = StreamOnlyLogger()

    @property
    def target_url(self):
        """

        **中文文档**

        获得数据所在的真实Url地址。
        """
        raise NotImplementedError

    def get_html(self):
        """

        **中文文档**

        从Url上获得Html。默认使用自动检测文本编码的crawlib中的
        spider.get_html(url)方法。不过有些网站需要用到cookie登陆。可以通过重写
        该方法来实现。
        """
        return spider.get_html(self.target_url)

    @staticmethod
    def fetch_data(html):
        """

        :param html: str, html
        :returns data: dict like data

        **中文文档**

        从得到的Html中解析出数据。支持对json友好的自典型数据。若对json不友好。
        则后面跟数据的操作需要用到pickle接口。
        """
        raise NotImplementedError

    @classmethod
    def crawl_one(cls, task):
        """

        **中文文档**

        抓取一个Url页面。
        """
        url = task.target_url
        cls.logger.info("crawl %s, %s left ..." % (url, task.left_counter))

        try:
            html = spider.get_html(url)
        except Exception as e:
            cls.logger.info("http error: %s" % e, 1)
            return

        try:
            data = task.fetch_data(html)
            if task.__class__.use_json:
                cls.objects(_id=task._id).update(
                    json=json.dumps(data), status=1)
            else:
                cls.objects(_id=task._id).update(
                    pickle=pickle.dumps(data), status=1)
            cls.logger.info("success! data: %s" % data, 1)
        except Exception as e:
            cls.logger.info("parse error: %s" % e, 1)
            return

    @classmethod
    def crawl_all(cls, use_json=True):
        """

        **中文文档**

        多线程爬虫。
        """
        # Add left counter
        counter = cls.objects(status=0).count()
        todo = list()
        for task in cls.objects(status=0):
            task.left_counter = counter
            counter -= 1
            todo.append(task)

        task.__class__.use_json = use_json

        pool = Pool(6)
        pool.map(cls.crawl_one, todo)

    @classmethod
    def set_finished(cls, _id):
        """Mark a task as 'finished'.
        """
        cls.objects(_id=_id).update(status=1)

    @classmethod
    def set_not_finished(cls, _id):
        """Mark a task as 'not finished'.
        """
        cls.objects(_id=_id).update(status=0)



#--- Use Case ---
# Crawl CVS store information from http://www.cvs.com/store-locator/
if __name__ == "__main__":
    from bs4 import BeautifulSoup
    from loggerFactory import SingleFileLogger

    domain = "http://www.cvs.com"
    connect = mongoengine.connect("cvs")  # use cvs database

    class CVSTask(Task):
        meta = {"abstract": True}

        @property
        def target_url(self):
            return domain + self._id

    class State(CVSTask):
        """A State Page.

        Example: http://www.cvs.com/store-locator/cvs-pharmacy-locations/Virginia
        """
        name = mongoengine.StringField()

        meta = {"collection": "state"}

        @staticmethod
        def fetch_data(html):
            data = list()
            soup = BeautifulSoup(html)

            div = soup.find("div", class_="states")
            for a in div.find_all("a"):
                data.append(
                    {"href": a["href"], "name": " ".join(a.text.strip().split(" ")[:-1])})
            return data

    class City(CVSTask):
        """A City Page.

        Example: http://www.cvs.com/store-locator/cvs-pharmacy-locations/Virginia/Alexandria
        """
        name = mongoengine.StringField()

        meta = {"collection": "city"}

        @staticmethod
        def fetch_data(html):
            data = list()
            soup = BeautifulSoup(html)

            div_ = soup.find("div", class_="stores-wrap")
            for div in div_.find_all("div", class_="each-store"):
                d = dict()
                try:
                    p = div.find("p", class_="store-address")
                    d["address"] = p.text.strip()
                except:
                    pass

                try:
                    p = div.find("p", class_="phone-number")
                    d["phone"] = p.text.strip().split("#")[0].strip()
                    d["store_number"] = p.text.strip().split("#")[-1].strip()
                except:
                    pass

                p = div.find("p", class_="directions-link")
                a = p.find("a")
                d["href"] = a["href"]
                if d:
                    data.append(d)

            return data

    import string
    alpha_letter = set(string.ascii_letters + string.digits + " ")

    def extract_alpha_letter(text):
        """Extract only alpha and digits characters from text.
        """
        chunks = list()
        for char in text:
            if char in alpha_letter:
                chunks.append(char)
        return "".join(chunks)

    class Store(CVSTask):
        """A Store Page.

        Example: http://www.cvs.com/store-locator/cvs-pharmacy-address/415+Monroe+Avenue-Alexandria-VA-22301/storeid=1410
        """
        address = mongoengine.StringField()
        phone = mongoengine.StringField()
        store_number = mongoengine.StringField()

        meta = {"collection": "store"}

        @staticmethod
        def fetch_data(html):
            data = dict()
            soup = BeautifulSoup(html)

            try:
                ul = soup.find("ul", id="serviceBadges")
                data["services"] = [
                    extract_alpha_letter(li.text.strip()).strip() for li in ul.find_all("li")]
            except:
                pass

            return data

    #--- Unittest ---
    def test_State_fetch_data():
        url = "http://www.cvs.com/store-locator/cvs-pharmacy-locations/Virginia"
        html = spider.get_html(url)
        data = State.fetch_data(html)
        js.pprint(data)

    # test_State_fetch_data()

    def test_City_fetch_data():
        url = "http://www.cvs.com/store-locator/cvs-pharmacy-locations/Virginia/Alexandria"
        html = spider.get_html(url)
        data = City.fetch_data(html)
        js.pprint(data)

    # test_City_fetch_data()

    def test_Store_fetch_data():
        url = "http://www.cvs.com/store-locator/cvs-pharmacy-address/415+Monroe+Avenue-Alexandria-VA-22301/storeid=1410"
        html = spider.get_html(url)
        data = Store.fetch_data(html)
        js.pprint(data)

    # test_Store_fetch_data()

    #--- Project Code ---
    def fill_state():
        """Put 51 states as entry points from
        http://www.cvs.com/store-locator/cvs-pharmacy-locations.
        """
        data = list()

        url = "http://www.cvs.com/store-locator/cvs-pharmacy-locations"
        html = spider.get_html(url)
        soup = BeautifulSoup(html)
        div = soup.find("div", class_="states")
        for a in div.find_all("a"):
            state = State(_id=a["href"], name=a.text.strip(), status=0)
            data.append(state)

        State.smart_insert(data)

    # fill_state()

    # State.crawl_all(use_json=True)

    def fill_city():
        """Take city data from state collection, and fill into city collection.
        """
        data = list()
        for state in State.objects(_id__ne=0):
            try:
                for d in json.loads(state.json):
                    city = City(_id=d["href"], name=d["name"], status=0)
                    data.append(city)
            except:
                pass
        City.smart_insert(data)

    # fill_city()

    # City.crawl_all(use_json=True)

    def fill_store():
        """Take city data from city collection, and fill into store collection.
        """
        data = list()
        for city in City.objects(_id__ne=0):
            try:
                for d in json.loads(city.json):
                    store = Store(_id=d["href"],
                                  address=d.get("address"),
                                  phone=d.get("phone"),
                                  store_number=d.get("store_number"),
                                  status=0)
                    data.append(store)
            except:
                pass
        Store.smart_insert(data)

#     fill_store()

#     Store.crawl_all(use_json=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
compress and decompress data using zlib.
"""

import sys
import types
import zlib
import base64
import pickle


PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str,
    integer_types = int,
    class_types = type,
    text_type = str
    binary_type = bytes

    MAXSIZE = sys.maxsize
else:
    string_types = basestring,
    integer_types = (int, long)
    class_types = (type, types.ClassType)
    text_type = unicode
    binary_type = str

__all__ = ["compress", "decompress"]


def _compress_obj(obj, level):
    """Compress object to bytes.
    """
    return zlib.compress(pickle.dumps(obj, protocol=2), level)


def _compress_str(s, level):
    """Compress str to bytes.
    """
    return zlib.compress(s.encode("utf-8"), level)


def _compress_bytes(b, level):
    """Compress bytes to bytes.
    """
    return zlib.compress(b, level)


def compress(obj, level=6, return_type="bytes"):
    """Compress anything to bytes or string.

    :params obj: 
    :params level: 
    :params return_type: if bytes, then return bytes; if str, then return
      base64.b64encode bytes in utf-8 string. 
    """
    if isinstance(obj, binary_type):
        b = zlib.compress(obj, level)
    elif isinstance(obj, string_types):
        b = zlib.compress(obj.encode("utf-8"), level)
    else:
        b = zlib.compress(pickle.dumps(obj, protocol=2), level)

    if return_type == "bytes":
        return b
    elif return_type == "str":
        return base64.b64encode(b).decode("utf-8")
    else:
        raise ValueError("'return_type' has to be one of 'bytes', 'str'!")


def decompress(obj, return_type="bytes"):
    """De-compress it to it's original.

    :params obj: Compressed object, could be bytes or str.
    :params return_type: if bytes, then return bytes; if str, then use 
      base64.b64decode; if obj, then use pickle.loads return an object. 
    """
    if isinstance(obj, binary_type):
        b = zlib.decompress(obj)
    elif isinstance(obj, string_types):
        b = zlib.decompress(base64.b64decode(obj.encode("utf-8")))
    else:
        raise TypeError("input cannot be anything other than str and bytes!")

    if return_type == "bytes":
        return b
    elif return_type == "str":
        return b.decode("utf-8")
    elif return_type == "obj":
        return pickle.loads(b)
    else:
        raise ValueError(
            "'return_type' has to be one of 'bytes', 'str' or 'obj'!")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module provides plenty of useful functions for iterable object manipulation.

**中文文档**

"""

import sys
import random
import collections
import itertools
if sys.version_info[0] == 2:
    from itertools import (
        ifilterfalse as filterfalse,
        izip_longest as zip_longest,
        izip as zip,
    )
    string_types = basestring,
else:  # in python3
    from itertools import filterfalse, zip_longest
    string_types = str,


def flatten(iterable):
    """Flatten one layer of nesting.

    Example::

        >>> list(flatten([[0, 1], [2, 3]])
        [0, 1, 2, 3]

        >>> list(flatten(["ab", "cd"])
        ["a", "b", "c", "d"]

    **中文文档**

    将二维列表压平成一维列表。
    """
    return itertools.chain.from_iterable(iterable)


def flatten_all(nested_iterable):
    """Flatten arbitrary depth of nesting. Good for unknown nesting structure
    iterable object.

    Example::

        >>> list(flatten_all([[1, 2], "abc", [3, ["x", "y", "z"]], 4]))
        [1, 2, "abc", 3, "x", "y", "z", 4]

    **中文文档**

    将任意维度的列表压平成一维列表。

    注: 使用hasattr(i, "__iter__")方法做是否是可循环对象的判断, 性能要高于其他
    任何方法, 例如: isinstance(i, collections.Iterable)
    """
    for item in nested_iterable:
        if hasattr(item, "__iter__") and not isinstance(item, string_types):
            for i in flatten_all(item):
                yield i
        else:
            yield item


def nth(iterable, n, default=None):
    """Returns the nth item or a default value.

    Example::

        >>> nth([0, 1, 2], 1)
        1

        >>> nth([0, 1, 2], 100)
        None

    **中文文档**

    取出一个可循环对象中的第n个元素。等效于list(iterable)[n], 但占用极小的内存。
    因为list(iterable)要将所有元素放在内存中并生成一个新列表。该方法常用语对于
    那些取index操作被改写了的可循环对象。
    """
    return next(itertools.islice(iterable, n, None), default)


def take(iterable, n):
    """Return first n items of the iterable as a list.

    Example::

        >>> take([0, 1, 2], 2)
        [0, 1]

    **中文文档**

    取出可循环对象中的前n个元素。等效于list(iterable)[:n], 但占用极小的内存。
    因为list(iterable)要将所有元素放在内存中并生成一个新列表。该方法常用语对于
    那些取index操作被改写了的可循环对象。
    """
    return list(itertools.islice(iterable, n))


def pull(iterable, n):
    """Return last n items of the iterable as a list.

    Example::

        >>> pull([0, 1, 2], 3)
        [1, 2]

    **中文文档**

    取出可循环对象中的最后n个元素。等效于list(iterable)[-n:], 但占用极小的内存。
    因为list(iterable)要将所有元素放在内存中并生成一个新列表。该方法常用语对于
    那些取index操作被改写了的可循环对象。
    """
    fifo = collections.deque(maxlen=n)
    for i in iterable:
        fifo.append(i)
    return list(fifo)


def shuffled(iterable):
    """Returns the shuffled iterable.

    Example::

        >>> shuffled([0, 1, 2])
        [2, 0, 1]

    **中文文档**

    打乱一个可循环对象中所有元素的顺序。并打包成列表返回。
    """
    return random.sample(iterable, len(iterable))


def grouper(iterable, n, fillvalue=None):
    """Collect data into fixed-length chunks or blocks.

    Example::

        >>> list(grouper(range(10), n=3, fillvalue=None))
        [(0, 1, 2), (3, 4, 5), (6, 7, 8), (9, None, None)]

    **中文文档**

    将一个序列按照尺寸n, 依次打包输出, 如果元素不够n的包, 则用 ``fillvalue`` 中的值填充。
    """
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def grouper_list(l, n):
    """Evenly divide list into fixed-length piece, no filled value if chunk
    size smaller than fixed-length.

    Example::

        >>> list(grouper(range(10), n=3)
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

    **中文文档**

    将一个列表按照尺寸n, 依次打包输出, 有多少输出多少, 并不强制填充包的大小到n。

    下列实现是按照性能从高到低进行排列的:

    - 方法1: 建立一个counter, 在向chunk中添加元素时, 同时将counter与n比较, 如果一致
      则yield。然后在最后将剩余的item视情况yield。
    - 方法2: 建立一个list, 每次添加一个元素, 并检查size。
    - 方法3: 调用grouper()函数, 然后对里面的None元素进行清理。
    """
    chunk = list()
    counter = 0
    for item in l:
        counter += 1
        chunk.append(item)
        if counter == n:
            yield chunk
            chunk = list()
            counter = 0
    if len(chunk) > 0:
        yield chunk


def grouper_dict(d, n, dict_type=dict):
    """Evenly divide dictionary into fixed-length piece, no filled value if
    chunk size smaller than fixed-length. Notice: dict is unordered in python,
    this method suits better for collections.OrdereDict

    Example::
        >>> d = {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5}
        >>> list(grouper_dict(d, 2)
        [{"a": 1, "b": 2}, {"c": 3, "d": 4}, {"e": 5}]

    **中文文档**

    将一个列表按照尺寸n, 依次打包输出, 有多少输出多少, 并不强制填充包的大小到n。
    """
    chunk = dict_type()
    counter = 0
    for k, v in d.items():
        counter += 1
        chunk[k] = v
        if counter == n:
            yield chunk
            chunk = dict_type()
            counter = 0
    if len(chunk) > 0:
        yield chunk

#--- Window ---


def running_window(iterable, size):
    """Generate n-size running window.

    Example::

        >>> for i in running_windows([1, 2, 3, 4, 5], size=3):
        ...     print(i)
        [1, 2, 3]
        [2, 3, 4]
        [3, 4, 5]

    **中文文档**

    简单滑窗函数。
    """
    if size > len(iterable):
        raise ValueError("size can not be greater than length of iterable.")

    fifo = collections.deque(maxlen=size)
    for i in iterable:
        fifo.append(i)
        if len(fifo) == size:
            yield list(fifo)


def cycle_running_window(iterable, size):
    """Generate n-size cycle running window.

    Example::

        >>> for i in running_windows([1, 2, 3, 4, 5], size=3):
        ...     print(i)
        [1, 2, 3]
        [2, 3, 4]
        [3, 4, 5]
        [4, 5, 1]
        [5, 1, 2]

    **中文文档**

    循环位移滑窗函数。
    """
    if size > len(iterable):
        raise ValueError("size can not be greater than length of iterable.")

    fifo = collections.deque(maxlen=size)
    cycle = itertools.cycle(iterable)
    counter = itertools.count(1)
    length = len(iterable)
    for i in cycle:
        fifo.append(i)
        if len(fifo) == size:
            yield list(fifo)
            if next(counter) == length:
                break

#--- Cycle ---


def cycle_slice(sliceable, start, end):
    """Given a list, return right hand cycle direction slice from start to end.

    Example::

        >>> array = [0, 1, 2, 3]
        >>> cycle_slice(array, 1, 3) # from array[1] to array[3]
        [1, 2]

        >>> cycle_slice(array, 3, 1) # from array[3] to array[1]
        [3, 0]

    **中文文档**

    """
    if type(sliceable) != list:
        sliceable = list(sliceable)
    length = len(sliceable)

    if length == 0:
        raise ValueError("sliceable cannot be empty!")
    start = start % length
    end = end % length

    if end > start:
        return sliceable[start:end]
    elif end <= start:
        return sliceable[start:] + sliceable[:end]


def cycle_dist(x, y, perimeter):
    """Find Distance between x, y by means of a n-length cycle.

    :param x:
    :param y:
    :param perimeter:

    Example:

        >>> cycle_dist(1, 23, 24) = 2
        >>> cycle_dist(5, 13, 24) = 8
        >>> cycle_dist(0.0, 2.4, 1.0) = 0.4
        >>> cycle_dist(0.0, 2.6, 1.0) = 0.4

    **中文文档**

    假设坐标轴是一个环, 计算两点之间在环上的最短距离。
    """
    dist = abs(x - y) % perimeter
    if dist > 0.5 * perimeter:
        dist = perimeter - dist
    return dist


#--- Shift ---
def cyclic_shift(array, dist):
    """

    :params array: list like iterable object
    :params dist: int

    Example::

        >>> cyclic_shift([0, 1, 2], 1)
        [2, 0, 1]

        >>> cyclic_shift([0, 1, 2], -1)
        [1, 2, 0]

    **中文文档**

    循环位移函数。
    """
    dist = dist % len(array)
    return array[-dist:] + array[:-dist]


def shift_and_trim(array, dist):
    """Shift and trim unneeded item.

    :params array: list like iterable object
    :params dist: int

    Example::

        >>> array = [0, 1, 2]

        >>> shift_and_trim(array, 0)
        [0, 1, 2]
        >>> shift_and_trim(array, 1)
        [0, 1]

        >>> shift_and_trim(array, -1)
        [1, 2]

        >>> shift_and_trim(array, 3)
        []

        >>> shift_and_trim(array, -3)
        []
    """
    length = len(array)
    if length == 0:
        return []

    if (dist >= length) or (dist <= -length):
        return []
    elif dist < 0:
        return array[-dist:]
    elif dist > 0:
        return array[:-dist]
    else:
        return list(array)


def shift_and_pad(array, dist, pad="__null__"):
    """Shift and pad with item.

    :params array: list like iterable object
    :params dist: int
    :params pad: any value

    Example::

        >>> array = [0, 1, 2]
        >>> shift_and_pad(array, 0)
        [0, 1, 2]

        >>> shift_and_pad(array, 1)
        [0, 0, 1]

        >>> shift_and_pad(array, -1)
        [1, 2, 2]

        >>> shift_and_pad(array, 3)
        [0, 0, 0]

        >>> shift_and_pad(array, -3)
        [2, 2, 2]

        >>> shift_and_pad(array, -1, None)
        [None, 0, 1]
    """
    length = len(array)
    if length == 0:
        return []

    if pad == "__null__":
        if dist > 0:
            padding_item = array[0]
        elif dist < 0:
            padding_item = array[-1]
        else:
            padding_item = None
    else:
        padding_item = pad

    if abs(dist) >= length:
        return length * [padding_item, ]
    elif dist == 0:
        return list(array)
    elif dist > 0:
        return [padding_item, ] * dist + array[:-dist]
    elif dist < 0:
        return array[-dist:] + [padding_item, ] * -dist
    else:  # Never get in this logic
        raise Exception


def size_of_generator(generator, memory_efficient=True):
    """Get number of items in a generator function.

    - memory_efficient = True, 3 times slower, but memory_efficient.
    - memory_efficient = False, faster, but cost more memory.

    **中文文档**

    计算一个生成器函数中的元素的个数。使用memory_efficient=True的方法可以避免将生成器中的
    所有元素放入内存, 但是速度稍慢于memory_efficient=False的方法。
    """
    if memory_efficient:
        counter = 0
        for _ in generator:
            counter += 1
        return counter
    else:
        return len(list(generator))

# Function


def difference(array, k=1):
    """Calculate l[n] - l[n-k]
    """
    if (len(array) - k) < 1:
        raise ValueError()
    if k < 0:
        raise ValueError("k has to be greater or equal than zero!")
    elif k == 0:
        return [i - i for i in array]
    else:
        return [j - i for i, j in zip(array[:-k], array[k:])]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
本模块提供了一些迷你的文本处理函数。

- 所有函数名都以 ``format_xxx`` 开头, 这样是为了防止在使用 
  ``from textformatter import *`` 时会污染命名空间。
"""

import string

FILENAME_FORBIDDEN_CHAR = r"""\/:*?|<'>"""
FUNCTION_WORD = set([
    "a", "an", "the",
    "and", "or", "not",
    "in", "on", "at",
    "am", "is", "are",
    "was", "were",
    "with", "within", "as", "of",
    "to", "from", "by",
    "than",
])
ALPHA_DIGITS = set(string.ascii_letters + string.digits)


def format_single_space_only(text):
    """Revise consecutive empty space to single space.

    Example::

        " I   feel    so  GOOD!" => "This is so GOOD!"

    **中文文档**

    确保文本中不会出现多余连续1次的空格。
    """
    return " ".join([word for word in text.strip().split(" ") if len(word) >= 1])


def format_title(text):
    """Capitalize first letter for each words except function words.

    Example::

        title = "Beautiful is Better than Ugly"

    **中文文档**

    将文本 "标题化", 即除了虚词, 每一个英文单词的第一个字母大写。
    """
    text = text.strip()
    # if empty string, return ""
    if len(text) == 0:
        return text
    else:
        text = text.lower()  # lower all char

        # Change to in single space format
        words = [word for word in text.strip().split(" ") if len(word) >= 1]

        # Capitalize all words except function word
        words_new = list()
        for word in words:
            if word not in FUNCTION_WORD:
                word = word[0].upper() + word[1:]
            words_new.append(word)

        # Make sure first word always be capitalized
        words_new[0] = words_new[0][0].upper() + words_new[0][1:]

        return " ".join(words_new)


def format_person_name(text):
    """Capitalize first letter for each part of the name.

    Example::

        person_name = "James Bond"

    **中文文档**

    将文本修改为人名格式。每个单词的第一个字母大写。
    """
    text = text.strip()
    if len(text) == 0:  # if empty string, return it
        return text
    else:
        text = text.lower()  # lower all char
        # delete redundant empty space
        words = [word for word in text.strip().split(" ") if len(word) >= 1]
        words = [word[0].upper() + word[1:] for word in words]
        return " ".join(words)


def format_filename(text):
    """Remove file-system forbidden character from file name.

    **中文文档**

    从文件名中移除文件系统中不允许的字符。 
    """
    for char in FILENAME_FORBIDDEN_CHAR:
        text = text.replace(char, "")
    return text


def format_CapitalizedWord(text):
    """

    Example::

        ThisIsVeryGood

    **中文文档**

    将文本格式化为各单词首字母大写, 拼接而成的长变量名。
    """
    text = text.strip()
    if len(text) == 0:  # if empty string, return it
        return text
    else:
        text = text.lower()  # lower all char
        # delete redundant empty space
        words = list()
        word = list()
        for char in text:
            if char in ALPHA_DIGITS:
                word.append(char)
            else:
                if len(word):
                    words.append("".join(word))
                    word = list()
        if len(word):
            words.append("".join(word))

        words = [word[0].upper() + word[1:] for word in words]
        return "".join(words)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module provides more ``marshmallow.fields.Fields``.
"""

from six import string_types
from marshmallow import fields, ValidationError


class AllowNoneField(fields.Field):
    """This field always allow None.

    **中文文档**

    默认允许None值。
    """

    def __init__(self, *args, **kwargs):
        kwargs["allow_none"] = True
        super(AllowNoneField, self).__init__(*args, **kwargs)


class AutoConvertField(AllowNoneField):
    """This field always call ``self.convert`` method before load and dump.

    **中文文档**

    在load和dump之前都自动调用covert函数进行数据预处理。常用于int, float或
    string这一类原生类。
    """

    def convert(self, value):
        raise NotImplementedError

    def _serialize(self, value, attr, obj):
        return self.convert(value)

    def _deserialize(self, value, attr, data):
        return self.convert(value)


class NonEmptyStringField(AutoConvertField):
    """A non empty string field.

    **中文文档**

    非空字符串。
    """

    def convert(self, value):
        if value is None:
            return None

        if isinstance(value, string_types):
            value = value.strip()
            if value:
                return value
            else:
                return None
        else:
            raise ValidationError("Not a string type.")


class TitleStringField(AutoConvertField):
    """Titlized string.

    **中文文档**

    标题格式的字符串。

    - 字符串
    - 前后都没有空格
    - 不得出现连续两个以上的空格
    - 每个单词首字母大写, 其他字母小写
    """

    def convert(self, value):
        if value is None:
            return None

        if isinstance(value, string_types):
            value = value.strip()
            if value:
                chunks = list()
                for s in [s.strip() for s in value.split(" ") if s.strip()]:
                    if str.isalpha(s):
                        s = s[0].upper() + s[1:].lower()
                        chunks.append(s)
                return " ".join(chunks)
            else:
                return None
        else:
            raise ValidationError("Not a string type")


class LowerStringField(AutoConvertField):
    """Lowercased string.

    **中文文档**

    - 前后没有空格
    - 全部小写
    """

    def convert(self, value):
        if value is None:
            return None

        if isinstance(value, string_types):
            value = value.strip().lower()
            if value:
                return value
            else:
                return None
        else:
            raise ValidationError("Not a string type")


class UpperStringField(AutoConvertField):
    """Uppercased string.

    **中文文档**

    - 前后没有空格
    - 全部小写
    """

    def convert(self, value):
        if value is None:
            return None

        if isinstance(value, string_types):
            value = value.strip().upper()
            if value:
                return value
            else:
                return None
        else:
            raise ValidationError("Not a string type")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A file compress utility module. You can easily programmatically add files
and directorys to zip archives. And compress arbitrary binary content.

- :func:`zip_a_folder`: add folder to archive.
- :func:`zip_everything_in_a_folder`: add everything in a folder to archive.
- :func:`zip_many_files`: Add many files to a zip archive.
- :func:`write_gzip`: Write binary content to gzip file.
- :func:`read_gzip`: Read binary content from gzip file.

**中文文档**

提供了若干个文件和数据压缩的快捷函数。

- :func:`zip_a_folder`: 将目录添加到压缩包。
- :func:`zip_everything_in_a_folder`: 将目录内的所有文件添加到压缩包。
- :func:`zip_many_files`: 将多个文件添加到压缩包。
- :func:`write_gzip`: 将二进制数据写入文件, 例如python pickle, bytes string。
- :func:`read_gzip`: 读取解压后的二进制数据内容。

注: python中zipfile包自带的ZipFile方法的用法如下:

基本用法::

    with ZipFile("filename.zip", "w") as f:
        f.write(path)

其中path是文件路径。 如果path是文件夹, 并不会将文件夹内所有的文件添加到压缩包中。

相对路径压缩:

比如你有一个路径 ``C:\download\readme.txt``, 如果当前路径是 ``C:\``,
而此时你将 ``readme.txt`` 添加到压缩包时则是在压缩包内添加一个: ``download\readme.txt``,
如果当前路径是 ``C:\download\``, 则在压缩包内添加的路径则是: ``readme.txt``
"""


from __future__ import print_function

import os
from zipfile import ZipFile


def zip_a_folder(src, dst):
    """Add a folder and everything inside to zip archive.

    Example::

        |---paper
            |--- algorithm.pdf
            |--- images
                |--- 1.jpg

        zip_a_folder("paper", "paper.zip")

        paper.zip
            |---paper
                |--- algorithm.pdf
                |--- images
                    |--- 1.jpg

    **中文文档**

    将整个文件夹添加到压缩包, 包括根目录本身。
    """
    if os.path.exists(dst):
        print("destination '%s' already exist." % dst)
        return

    src, dst = os.path.abspath(src), os.path.abspath(dst)
    cwd = os.getcwd()
    todo = list()

    dirname, basename = os.path.split(src)
    os.chdir(dirname)
    for dirname, _, fnamelist in os.walk(basename):
        for fname in fnamelist:
            newname = os.path.join(dirname, fname)
            todo.append(newname)

    with ZipFile(dst, "w") as f:
        for newname in todo:
            f.write(newname)

    os.chdir(cwd)


def zip_everything_in_a_folder(src, dst):
    """Add everything in a folder except the root folder it self to zip archive.

    Example::

        |---paper
            |--- algorithm.pdf
            |--- images
                |--- 1.jpg

        zip_everything_in_folder("paper", "paper.zip")

        paper.zip
            |--- algorithm.pdf
            |--- images
                |--- 1.jpg

    **中文文档**

    将目录内部的所有文件添加到压缩包, 不包括根目录本身。
    """
    if os.path.exists(dst):
        print("destination '%s' already exist." % dst)
        return

    src, dst = os.path.abspath(src), os.path.abspath(dst)
    cwd = os.getcwd()
    todo = list()

    os.chdir(src)
    for dirname, _, fnamelist in os.walk(os.getcwd()):
        for fname in fnamelist:
            newname = os.path.relpath(os.path.join(dirname, fname), src)
            todo.append(newname)

    with ZipFile(dst, "w") as f:
        for newname in todo:
            f.write(newname)

    os.chdir(cwd)


def zip_many_files(list_of_abspath, dst):
    """Add many files to a zip archive.

    **中文文档**

    将一系列的文件压缩到一个压缩包中, 若有重复的文件名, 在zip中保留所有的副本。
    """
    if os.path.exists(dst):
        print("destination '%s' already exist." % dst)
        return

    base_dir = os.getcwd()

    with ZipFile(dst, "w") as f:
        for abspath in list_of_abspath:
            dirname, basename = os.path.split(abspath)
            os.chdir(dirname)
            f.write(basename)

    os.chdir(base_dir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module extend the power of mongoengine.
"""

import math
import random
import mongoengine
from collections import OrderedDict
from copy import deepcopy


def grouper_list(l, n):
    """Evenly divide list into fixed-length piece, no filled value if chunk
    size smaller than fixed-length.

    Example::

        >>> list(grouper(range(10), n=3)
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

    **中文文档**

    将一个列表按照尺寸n, 依次打包输出, 有多少输出多少, 并不强制填充包的大小到n。

    下列实现是按照性能从高到低进行排列的:

    - 方法1: 建立一个counter, 在向chunk中添加元素时, 同时将counter与n比较, 如果一致
      则yield。然后在最后将剩余的item视情况yield。
    - 方法2: 建立一个list, 每次添加一个元素, 并检查size。
    - 方法3: 调用grouper()函数, 然后对里面的None元素进行清理。
    """
    chunk = list()
    counter = 0
    for item in l:
        counter += 1
        chunk.append(item)
        if counter == n:
            yield chunk
            chunk = list()
            counter = 0
    if len(chunk) > 0:
        yield chunk


class ExtendedDocument(mongoengine.Document):

    """Provide `mongoengine.Document <http://docs.mongoengine.org/apireference.html#mongoengine.Document>`_
    more utility methods.

    **中文文档**

    为默认的 ``mongoengine.Document`` 提供了更多的便捷方法。
    """
    meta = {
        "abstract": True,
    }

    def keys(self):
        """Convert to field list.
        """
        try:
            return list(self._fields_ordered)
        except AttributeError:
            raise TypeError("It does not inherit from mongoengine.Document!")

    def values(self):
        """Convert to field value list.
        """
        return [self._data.get(attr) for attr in self.keys()]

    def items(self):
        """Convert to field and value pair list.
        """
        return [(attr, self._data.get(attr)) for attr in self.keys()]

    def to_tuple(self):
        """Convert to field tuple.
        """
        return self._fields_ordered

    def to_list(self):
        """Convert to field list.
        """
        return self.keys()

    def to_dict(self):
        """Convert to dict.
        """
        return dict(self.items())

    def to_OrderedDict(self):
        """Convert to OrderedDict.
        """
        return OrderedDict(self.items())

    def __repr__(self):
        kwargs = list()
        for attr, value in self.items():
            kwargs.append("%s=%r" % (attr, value))
        return "%s(%s)" % (self.__class__.__name__, ", ".join(kwargs))

    def __str__(self):
        return self.__repr__()

    def absorb(self, other):
        """For attributes of others that value is not None, assign it to self.

        **中文文档**

        将另一个文档中的数据更新到本条文档。当且仅当数据值不为None时。
        """
        if not isinstance(other, self.__class__):
            raise TypeError

        for attr, value in other.items():
            if value is not None:
                setattr(self, attr, deepcopy(value))

    def revise(self, data):
        """Revise attributes value with dictionary data.

        **中文文档**

        将一个字典中的数据更新到本条文档。当且仅当数据值不为None时。
        """
        if not isinstance(data, dict):
            raise TypeError

        for key, value in data.items():
            if value is not None:
                setattr(self, key, deepcopy(value))

    @classmethod
    def collection(cls):
        """Get pymongo Collection instance.

        **中文文档**

        获得pymongo.Collection的实例。
        """
        return cls._get_collection()

    @classmethod
    def col(cls):
        """Get pymongo Collection instance.

        **中文文档**

        获得pymongo.Collection的实例。
        """
        return cls._get_collection()

    @classmethod
    def smart_insert(cls, data, minimal_size=5):
        """An optimized Insert strategy.

        **中文文档**

        在Insert中, 如果已经预知不会出现IntegrityError, 那么使用Bulk Insert的速度要
        远远快于逐条Insert。而如果无法预知, 那么我们采用如下策略:

        1. 尝试Bulk Insert, Bulk Insert由于在结束前不Commit, 所以速度很快。
        2. 如果失败了, 那么对数据的条数开平方根, 进行分包, 然后对每个包重复该逻辑。
        3. 若还是尝试失败, 则继续分包, 当分包的大小小于一定数量时, 则使用逐条插入。
          直到成功为止。

        该Insert策略在内存上需要额外的 sqrt(nbytes) 的开销, 跟原数据相比体积很小。
        但时间上是各种情况下平均最优的。
        """
        if isinstance(data, list):
            # 首先进行尝试bulk insert
            try:
                cls.objects.insert(data)
            # 失败了
            except mongoengine.NotUniqueError:
                # 分析数据量
                n = len(data)
                # 如果数据条数多于一定数量
                if n >= minimal_size ** 2:
                    # 则进行分包
                    n_chunk = math.floor(math.sqrt(n))
                    for chunk in grouper_list(data, n_chunk):
                        cls.smart_insert(chunk, minimal_size)
                # 否则则一条条地逐条插入
                else:
                    for document in data:
                        try:
                            cls.objects.insert(document)
                        except mongoengine.NotUniqueError:
                            pass
        else:
            try:
                cls.objects.insert(data)
            except mongoengine.NotUniqueError:
                pass

    @classmethod
    def by_id(cls, _id):
        """Get one instance by _id.

        **中文文档**

        根据_id, 返回一条文档。
        """
        return cls.objects(__raw__={"_id": _id}).get()

    @classmethod
    def by_filter(cls, filters):
        """Filter objects by pymongo dict query.

        **中文文档**

        使用pymongo的API进行query。
        """
        return cls.objects(__raw__=filters)

    @classmethod
    def random_sample(cls, filters=None, n=5):
        """

        **中文文档**

        从collection中随机选择 ``n`` 个样本。
        """
        data = list()

        id_field = cls._meta["id_field"]

        if filters is None:
            pipeline = [{"$sample": {"size": n}}]
            col = cls.col()

            if id_field == "_id":
                for doc in col.aggregate(pipeline):
                    obj = cls(**doc)
                    data.append(obj)

            else:
                for doc in col.aggregate(pipeline):
                    doc[id_field] = doc["_id"]
                    del doc["_id"]
                    obj = cls(**doc)
                    data.append(obj)
        else:
            if id_field == "_id":
                for doc in random.sample(list(cls.col().find(filters)), n):
                    obj = cls(**doc)
                    data.append(obj)
            else:
                filters["_id"] = filters[id_field]
                del filters[id_field]
                for doc in random.sample(list(cls.col().find(filters)), n):
                    doc[id_field] = doc["_id"]
                    del doc["_id"]
                    obj = cls(**doc)
                    data.append(obj)

        return data


#--- Unittest ---
class User(ExtendedDocument):
    user_id = mongoengine.IntField(primary_key=True)
    name = mongoengine.StringField()

    meta = {"collection": "user"}


if __name__ == "__main__":
    import mongoengine

    mongoengine.connect()


def test_smart_insert():
    import time
    import random

    # Smart Insert
    User.objects.delete()

    users = set([User(user_id=random.randint(1, 10000)) for i in range(20)])
    User.objects.insert(users)
    assert 15 <= User.objects.count() <= 20

    users = [User(user_id=i) for i in range(1, 1 + 10000)]

    st = time.clock()
    User.smart_insert(users)
    elapse1 = time.clock() - st

    assert User.objects.count() == 10000  # after smart insert, we got 10000 doc

    # Regular Insert
    User.objects.delete()

    users = set([User(user_id=random.randint(1, 10000)) for i in range(20)])
    User.objects.insert(users)
    assert 15 <= User.objects.count() <= 20

    users = [User(user_id=i) for i in range(1, 1 + 10000)]

    st = time.clock()
    for user in users:
        try:
            user.save()
        except:
            pass
    elapse2 = time.clock() - st

    assert User.objects.count() == 10000  # after regular insert, we got 10000 doc

    assert elapse1 <= elapse2


def test_query():
    User(user_id=1, name="Jack").save()
    User(user_id=2, name="Tom").save()

    assert User.by_id(1).name == "Jack"
    assert User.by_filter({"_id": 2})[:][0].name == "Tom"


def test_random_sample():
    User.col().remove({})
    for i in range(1000):
        user = User(user_id=i)
        user.save()

    assert len(User.random_sample(n=3)) == 3
    for user in User.random_sample(filters={"user_id": {"$gte": 500}}, n=3):
        assert user.user_id >= 500


if __name__ == "__main__":
    #
    test_smart_insert()
    test_query()
    test_random_sample()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

**中文文档**

matplotmate提供了一套函数式的命令, 用于时间序列数据的简单绘图。适画用于在R&D中
用于快速观测的图。自定义程度较低, 但是默认设置非常适合研究。
"""

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.dates import YearLocator, MonthLocator, WeekdayLocator, DayLocator
from matplotlib.dates import HourLocator, MinuteLocator
from matplotlib.dates import DateFormatter
from matplotlib.dates import MO, TU, WE, TH, FR, SA, SU
from itertools import cycle

CYCOl = cycle("brgcmyk")


def one_day_formatter():
    hours = HourLocator(range(24))
    hoursFmt = DateFormatter("%H:%M")
    minutes = MinuteLocator([30, ])
    minutesFmt = DateFormatter("%M")
    return hours, hoursFmt, minutes, minutesFmt


def one_week_formatter():
    days = DayLocator(range(365))
    daysFmt = DateFormatter("%m-%d")
    hours = HourLocator([3, 6, 9, 12, 15, 18, 21])
    hoursFmt = DateFormatter("%H")
    return days, daysFmt, hours, hoursFmt


def one_month_formatter():
    days = DayLocator(range(365))
    daysFmt = DateFormatter("%m-%d")
    return days, daysFmt, None, None


def one_quarter_formatter():
    months = MonthLocator(range(1, 13))
    monthsFmt = DateFormatter("%Y-%m")
    days = DayLocator([7, 14, 21, 28])
    daysFmt = DateFormatter("%dth")
    return months, monthsFmt, days, daysFmt


def one_year_formatter():
    months = MonthLocator(range(1, 13))
    monthsFmt = DateFormatter("%Y-%m")
    days = DayLocator([15])
    daysFmt = DateFormatter("%d")
    return months, monthsFmt, days, daysFmt


def format_x_tick(axis,
                  major_locator=None,
                  major_formatter=None,
                  minor_locator=None,
                  minor_formatter=None):
    """Set x axis's format.

    This method is designed for time axis.

    **中文文档**

    设置X轴格式。
    """
    if major_locator:
        axis.xaxis.set_major_locator(major_locator)
    if major_formatter:
        axis.xaxis.set_major_formatter(major_formatter)
    if minor_locator:
        axis.xaxis.set_minor_locator(minor_locator)
    if minor_formatter:
        axis.xaxis.set_minor_formatter(minor_formatter)

    axis.autoscale_view()
    plt.setp(axis.xaxis.get_majorticklabels(), rotation=90)
    plt.setp(axis.xaxis.get_minorticklabels(), rotation=90)
    axis.grid()


def set_title_xlabel_ylabel(axis, title, xlabel=None, ylabel=None):
    """

    **中文文档**

    设置标题, x坐标和y坐标的单位标签。
    """
    if title:
        axis.set_title(title)

    if xlabel:
        axis.set_xlabel(xlabel)
    else:
        axis.set_xlabel("Time")

    if ylabel:
        axis.set_ylabel(ylabel)
    else:
        axis.set_ylabel("Value")


def set_ylim(axis, y_min, y_max):
    """set min and max value of y axis.

    **中文文档**

    设置y坐标的最大最小值。
    """
    axis.set_ylim([y_min, y_max])


def set_legend(axis, lines, legend):
    """Set line legend.

    **中文文档**

    设置图例。
    """
    try:
        if legend:
            axis.legend(lines, legend)
    except Exception as e:
        raise ValueError("invalid 'legend', Error: %s" % e)


def get_max(array):
    """Get maximum value of an array. Automatically ignore invalid data.

    **中文文档**

    获得最大值。
    """
    largest = -np.inf
    for i in array:
        try:
            if i > largest:
                largest = i
        except:
            pass
    if np.isinf(largest):
        raise ValueError("there's no numeric value in array!")
    else:
        return largest


def get_min(array):
    """Get minimum value of an array. Automatically ignore invalid data.

    **中文文档**

    获得最小值。
    """
    smallest = np.inf
    for i in array:
        try:
            if i < smallest:
                smallest = i
        except:
            pass
    if np.isinf(smallest):
        raise ValueError("there's no numeric value in array!")
    else:
        return smallest


def get_yAxis_limit(y, lower=0.05, upper=0.2):
    """Find optimal y_min and y_max that guarantee enough space for legend and 
    plot.

    **中文文档**

    计算y坐标轴的最小和最大坐标。

    :params lower: ymin为 y的最小值再减去gap的一定倍率
    :params upper: ymax为 y的最大值再加上gap的一定倍率    
    """
    smallest = get_min(y)
    largest = get_max(y)
    gap = largest - smallest
    if gap >= 0.000001:
        y_min = smallest - lower * gap
        y_max = largest + upper * gap
    else:
        y_min = smallest - lower * abs(smallest)
        y_max = largest + upper * abs(largest)
    return y_min, y_max


def create_figure(width=20, height=10):
    """Create a figure instance.

    :params width: figure width
    :params height: figure height
    """
    figure = plt.figure(figsize=(width, height))
    axis = figure.add_subplot(1, 1, 1)
    return figure, axis


def preprocess_x_y(x, y):
    """Preprocess x, y input data. Returns list of list style.

    **中文文档**

    预处理输入的x, y数据。
    """
    def is_iterable_slicable(a):
        if hasattr(a, "__iter__") and hasattr(a, "__getitem__"):
            return True
        else:
            return False

    if is_iterable_slicable(x):
        if is_iterable_slicable(x[0]):
            return x, y
        else:
            return (x,), (y,)
    else:
        raise ValueError("invalid input!")


def plot_time_series(x, y,
                     linewidth=1, linestyle="-",
                     xlabel=None, ylabel=None,
                     x_major_locattor=None, x_major_formatter=None,
                     x_minor_locattor=None, x_minor_formatter=None,
                     title=None, legend=None, cycol=None):
    """
    :param x: time array or tuple
    :param y: time array or tuple

    **中文文档**

    为时间序列数据画图。
    """
    x, y = preprocess_x_y(x, y)
    if cycol is None:
        cycol = cycle("brgcmyk")
    plt.close("all")
    figure, axis = create_figure()

    lines = list()
    for time, value in zip(x, y):
        lines.append(axis.plot(time, value,
                               lw=linewidth, ls=linestyle, c=next(cycol))[0])

    format_x_tick(axis,
                  x_major_locattor, x_major_formatter,
                  x_minor_locattor, x_minor_formatter,)

    y_min, y_max = get_yAxis_limit(
        np.array(y).flatten(),
        lower=0.05,
        upper=0.1 * len(x),
    )
    set_ylim(axis, y_min, y_max)

    set_title_xlabel_ylabel(axis, title, xlabel, ylabel)

    set_legend(axis, lines, legend)

    return plt


def plot_one_day(x, y,
                 linewidth=1, linestyle="-",
                 xlabel=None, ylabel=None, title=None, legend=None, cycol=None):
    """
    """
    (
        x_major_locattor,
        x_major_formatter,
        x_minor_locattor,
        x_minor_formatter,
    ) = one_day_formatter()
    return plot_time_series(x, y, linewidth, linestyle, xlabel, ylabel,
                            x_major_locattor, x_major_formatter,
                            x_minor_locattor, x_minor_formatter,
                            title, legend, cycol,
                            )


def plot_one_week(x, y,
                  linewidth=1, linestyle="-",
                  xlabel=None, ylabel=None, title=None, legend=None, cycol=None):
    """
    """
    (
        x_major_locattor,
        x_major_formatter,
        x_minor_locattor,
        x_minor_formatter,
    ) = one_week_formatter()
    return plot_time_series(x, y, linewidth, linestyle, xlabel, ylabel,
                            x_major_locattor, x_major_formatter,
                            x_minor_locattor, x_minor_formatter,
                            title, legend, cycol,
                            )


def plot_one_month(x, y,
                   linewidth=1, linestyle="-",
                   xlabel=None, ylabel=None, title=None, legend=None, cycol=None):
    """
    """
    (
        x_major_locattor,
        x_major_formatter,
        x_minor_locattor,
        x_minor_formatter,
    ) = one_month_formatter()
    return plot_time_series(x, y, linewidth, linestyle, xlabel, ylabel,
                            x_major_locattor, x_major_formatter,
                            x_minor_locattor, x_minor_formatter,
                            title, legend, cycol,
                            )


def plot_one_quarter(x, y,
                     linewidth=1, linestyle="-",
                     xlabel=None, ylabel=None, title=None, legend=None, cycol=None):
    """
    """
    (
        x_major_locattor,
        x_major_formatter,
        x_minor_locattor,
        x_minor_formatter,
    ) = one_quarter_formatter()
    return plot_time_series(x, y, linewidth, linestyle, xlabel, ylabel,
                            x_major_locattor, x_major_formatter,
                            x_minor_locattor, x_minor_formatter,
                            title, legend, cycol,
                            )


def plot_one_year(x, y,
                  linewidth=1, linestyle="-",
                  xlabel=None, ylabel=None, title=None, legend=None, cycol=None):
    """
    """
    (
        x_major_locattor,
        x_major_formatter,
        x_minor_locattor,
        x_minor_formatter,
    ) = one_year_formatter()
    return plot_time_series(x, y, linewidth, linestyle, xlabel, ylabel,
                            x_major_locattor, x_major_formatter,
                            x_minor_locattor, x_minor_formatter,
                            title, legend,
                            )

#--- Twin Axis ---


def plot_two_scales(x1, y1, x2, y2,
                    linewidth=1, linestyle="-",
                    xlabel=None, ylabel1=None, ylabel2=None,
                    x_major_locattor=None, x_major_formatter=None,
                    x_minor_locattor=None, x_minor_formatter=None,
                    title=None, legend=None, cycol=None):
    """

    **中文文档**

    将两种不同尺度的线画在一张图上。
    """
    x1, y1 = preprocess_x_y(x1, y1)
    x2, y2 = preprocess_x_y(x2, y2)

    if cycol is None:
        cycol = cycle("brgcmyk")

    plt.close("all")
    figure, axis1 = create_figure()
    axis2 = axis1.twinx()

    lines = list()
    for time, value in zip(x1, y1):
        lines.append(axis1.plot(time, value,
                                lw=linewidth, ls=linestyle, c=next(cycol))[0])

    for time, value in zip(x2, y2):
        lines.append(axis2.plot(time, value,
                                lw=linewidth, ls=linestyle, c=next(cycol))[0])

    format_x_tick(axis1,
                  x_major_locattor, x_major_formatter,
                  x_minor_locattor, x_minor_formatter,)

    y1_min, y1_max = get_yAxis_limit(
        np.array(y1).flatten(),
        lower=0.05,
        upper=0.1 * len(y1),
    )
    set_ylim(axis1, y1_min, y1_max)

    y2_min, y2_max = get_yAxis_limit(
        np.array(y2).flatten(),
        lower=0.05,
        upper=0.1 * len(y2),
    )
    set_ylim(axis2, y2_min, y2_max)

    set_title_xlabel_ylabel(axis1, title, xlabel, ylabel1)

    set_title_xlabel_ylabel(axis2, title, xlabel, ylabel2)

    set_legend(axis1, lines, legend)

    return plt


#--- Unittest ---
if __name__ == "__main__":
    from pytest import raises, approx
    from rolex import rolex

    def test_get_max_get_min():
        assert get_max(np.array([0, 1, 2])) == 2
        assert get_max(np.array([None, 1, 2])) == 2
        with raises(ValueError):
            get_max(np.array([None, None, None]))

        assert get_min(np.array([0, 1, 2])) == 0
        assert get_min(np.array([None, 1, 2])) == 1
        with raises(ValueError):
            get_min(np.array([None, None, None]))

    test_get_max_get_min()

    def test_get_yAxis_limit():
        assert get_yAxis_limit([0, 1], lower=0.05, upper=0.2) \
            == approx([-0.05, 1.2])

    test_get_yAxis_limit()

    def test_preprocess_x_y():
        l = [1, 2, 3]
        t = (1, 2, 3)
        x, y = preprocess_x_y(l, l)
        assert x == (l, )

        x, y = preprocess_x_y(t, t)
        assert x == (t, )

        x, y = preprocess_x_y([t, t], [t, t])
        assert x == [t, t]

        x, y = preprocess_x_y((t, t), (t, t))
        assert x == (t, t)

    test_preprocess_x_y()

    def test_plot_one_day():
        x = rolex.time_series(
            "2016-01-01 00:00:00", "2016-01-01 23:59:59", freq="10min")
        y1 = np.random.random(len(x))
        y2 = np.random.random(len(x))
        plot_one_day((x, x), (y1, y2),
                     xlabel="Time", ylabel="Stock Value", title="Stock Trends",
                     legend=["Stock 1", "Stock 2"]).show()

    test_plot_one_day()

    def test_plot_one_week():
        x = rolex.time_series(
            "2016-01-01 00:00:00", "2016-01-07 23:59:59", freq="3hour")
        y1 = np.random.random(len(x))
        y2 = np.random.random(len(x))
        plot_one_week((x, x), (y1, y2),
                      xlabel="Time", ylabel="Stock Value", title="Stock Trends",
                      legend=["Stock 1", "Stock 2"]).show()

    test_plot_one_week()

    def test_plot_one_month():
        x = rolex.time_series(
            "2016-01-01 00:00:00", "2016-01-31 23:59:59", freq="12hour")
        y1 = np.random.random(len(x))
        y2 = np.random.random(len(x))
        plot_one_month((x, x), (y1, y2),
                       xlabel="Time", ylabel="Stock Value", title="Stock Trends",
                       legend=["Stock 1", "Stock 2"]).show()

    test_plot_one_month()

    def test_plot_one_quarter():
        x = rolex.time_series(
            "2016-01-01 00:00:00", "2016-03-31 23:59:59", freq="1day")
        y1 = np.random.random(len(x))
        y2 = np.random.random(len(x))
        plot_one_quarter((x, x), (y1, y2),
                         xlabel="Time", ylabel="Stock Value", title="Stock Trends",
                         legend=["Stock 1", "Stock 2"]).show()

    test_plot_one_quarter()

    def test_plot_one_year():
        x = rolex.time_series(
            "2016-01-01 00:00:00", "2016-12-31 23:59:59", freq="7day")
        y1 = np.random.random(len(x))
        y2 = np.random.random(len(x))
        plot_one_year((x, x), (y1, y2),
                      xlabel="Time", ylabel="Stock Value", title="Stock Trends",
                      legend=["Stock 1", "Stock 2"]).show()

    test_plot_one_year()

    def test_plot_two_scales():
        x = rolex.time_series(
            "2016-01-01 00:00:00", "2016-01-01 23:59:59", freq="10min")
        y11 = np.random.random(len(x))
        y12 = np.random.random(len(x))
        y21 = np.random.random(len(x)) * 10
        y22 = np.random.random(len(x)) * 10
        plot_two_scales((x, x), (y11, y12), (x, x), (y21, y22),
                        xlabel="Time", ylabel1="Interests Value", ylabel2="Stock Value", title="Stock Trends",
                        legend=["Interests1", "Interests2", "Stock1", "Stock2"]).show()

    test_plot_two_scales()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module provides some easy function to generate random text from built-in 
templates.

- :func:`rand_str`: fixed-length string
- :func:`rand_hexstr`: fixed-length hex string
- :func:`rand_pwd`: random password
- :func:`rand_phone`: random phone number
- :func:`rand_ssn`: random ssn
- :func:`rand_email`: random email
"""

import random
import string

CHARSET_ALPHA_DIGITS = string.ascii_letters + string.digits
CHARSET_PASSWORD = CHARSET_ALPHA_DIGITS + "!@#$%^&*()"
CHARSET_HEXSTR_LOWER = "0123456789abcdef"
CHARSET_HEXSTR_UPPER = CHARSET_HEXSTR_LOWER.upper()
DOMAIN_SURFIX = ["com", "net", "org", "edu"]


def rand_str(length, allowed=CHARSET_ALPHA_DIGITS):
    """Generate fixed-length random string from your allowed character pool.

    :param length: total length of this string.
    :param allowed: allowed charset.

    Example::

        >>> import string
        >>> rand_str(32)
        H6ExQPNLzb4Vp3YZtfpyzLNPFwdfnwz6
    """
    res = list()
    for _ in range(length):
        res.append(random.choice(allowed))
    return "".join(res)


def rand_hexstr(length, lower=True):
    """Gererate fixed-length random hexstring, usually for md5.

    :param length: total length of this string.
    :param lower: use lower case or upper case.
    """
    if lower:
        return rand_str(length, allowed=CHARSET_HEXSTR_LOWER)
    else:
        return rand_str(length, allowed=CHARSET_HEXSTR_UPPER)


def rand_alphastr(length, lower=True, upper=True):
    """Generate fixed-length random alpha only string.
    """
    if lower is True and upper is True:
        return rand_str(length, allowed=string.ascii_letters)
    if lower is True and upper is False:
        return rand_str(length, allowed=string.ascii_lowercase)
    if lower is False and upper is True:
        return rand_str(length, allowed=string.ascii_uppercase)
    else:
        raise


def rand_pwd(length):
    """Random Internet password.

    Example::

        >>> rand_pwd(12)
        TlhM$^jzculH
    """
    return rand_str(length, CHARSET_PASSWORD)


def rand_phone():
    """Random US phone number. (10 digits)

    Example::

        >>> rand_phone()
        (306)-746-6690
    """
    return "(%s)-%s-%s" % (rand_str(3, string.digits),
                           rand_str(3, string.digits),
                           rand_str(3, string.digits))


def rand_ssn():
    """Random SSN. (9 digits)

    Example::

        >>> rand_ssn()
        295-50-0178
    """
    return "%s-%s-%s" % (rand_str(3, string.digits),
                         rand_str(2, string.digits),
                         rand_str(4, string.digits))


def rand_email():
    """Random email.

    Usage Example::

        >>> rand_email()
        Z4Lljcbdw7m@npa.net
    """
    name = rand_str(random.randint(4, 14), string.ascii_lowercase) + \
        rand_str(random.randint(1, 4), string.digits)
    domain = rand_str(random.randint(2, 10), string.ascii_lowercase)
    surfix = random.choice(DOMAIN_SURFIX)
    return "%s@%s.%s" % (name, domain, surfix)


def rand_article(num_p=(4, 10), num_s=(2, 15), num_w=(5, 40)):
    """Random article text.

    Example::

        >>> rand_article()
        ...
    """
    article = list()
    for _ in range(random.randint(*num_p)):
        p = list()
        for _ in range(random.randint(*num_s)):
            s = list()
            for _ in range(random.randint(*num_w)):
                s.append(
                    rand_str(random.randint(1, 15), string.ascii_lowercase))
            p.append(" ".join(s))
        article.append(". ".join(p))
    return "\n\n".join(article)


#--- Just a simple API call for faker ---
try:
    import faker

    class SimpleFaker(object):
        def __init__(self, locale="en_US"):
            self.fake = faker.Factory.create(locale)

    simple_faker = SimpleFaker()
except:
    pass


def set_locale(locale):
    simple_faker.fake = faker.Factory.create(locale=locale)


def first_name():
    return simple_faker.fake.first_name()


def last_name():
    return simple_faker.fake.last_name()


def name():
    return simple_faker.fake.name()


def address():
    return simple_faker.fake.address()


def company():
    return simple_faker.fake.company()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import heapq
from math import radians, cos, sin, asin, sqrt
from sqlalchemy import create_engine, MetaData, Table, Column, Index
from sqlalchemy import String, Float, PickleType
from sqlalchemy import select, and_, func


AVG_EARTH_RADIUS = 6371  # in km


def great_circle(point1, point2, miles=True):
    """ Calculate the great-circle distance bewteen two points on the Earth surface.
    :input: two 2-tuples, containing the latitude and longitude of each point
    in decimal degrees.
    Example: great_circle((45.7597, 4.8422), (48.8567, 2.3508))
    :output: Returns the distance bewteen the two points.
    The default unit is kilometers. Miles can be returned
    if the ``miles`` parameter is set to True.
    """
    # unpack latitude/longitude
    lat1, lng1 = point1
    lat2, lng2 = point2

    # convert all latitudes/longitudes from decimal degrees to radians
    lat1, lng1, lat2, lng2 = list(map(radians, [lat1, lng1, lat2, lng2]))

    # calculate haversine
    lat = lat2 - lat1
    lng = lng2 - lng1
    d = sin(lat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(lng / 2) ** 2
    h = 2 * AVG_EARTH_RADIUS * asin(sqrt(d))
    if miles:
        return h * 0.621371  # in miles
    else:
        return h  # in kilometers


class GeoSearchEngine(object):
    def __init__(self, name="point", database=":memory:"):
        self.engine = create_engine("sqlite:///%s" % database)
        self.metadata = MetaData()
        self.t_point = Table(name, self.metadata,
                             Column("id", String),
                             Column("lat", Float),
                             Column("lng", Float),
                             Column("data", PickleType),
                             )

    def train(self, data, key_id, key_lat, key_lng, clear_old=True):
        """Feed data into database.
        """
        engine, t_point = self.engine, self.t_point
        if clear_old:
            try:
                t_point.drop(engine)
            except:
                pass
        t_point.create(engine)

        table_data = list()
        for record in data:
            id = key_id(record)
            lat = key_lat(record)
            lng = key_lng(record)
            row = {"id": id, "lat": lat, "lng": lng, "data": record}
            table_data.append(row)

        ins = t_point.insert()
        engine.execute(ins, table_data)

        index = Index('idx_lat_lng', t_point.c.lat, t_point.c.lng)
        index.create(engine)

    def find_n_nearest(self, lat, lng, n=5, radius=None):
        """Find n nearest point within certain distance from a point.

        :param lat: latitude of center point.
        :param lng: longitude of center point. 
        :param n: max number of record to return.
        :param radius: only search point within ``radius`` distance.

        **中文文档**
        """
        engine, t_point = self.engine, self.t_point
        if radius:
            # Use a simple box filter to minimize candidates
            # Define latitude longitude boundary
            dist_btwn_lat_deg = 69.172
            dist_btwn_lon_deg = cos(lat) * 69.172
            lat_degr_rad = abs(radius * 1.05 / dist_btwn_lat_deg)
            lon_degr_rad = abs(radius * 1.05 / dist_btwn_lon_deg)

            lat_lower = lat - lat_degr_rad
            lat_upper = lat + lat_degr_rad
            lng_lower = lng - lon_degr_rad
            lng_upper = lng + lon_degr_rad

            filters = [
                t_point.c.lat >= lat_lower,
                t_point.c.lat <= lat_upper,
                t_point.c.lat >= lng_lower,
                t_point.c.lat >= lng_upper,
            ]
        else:
            radius = 999999.9
            filters = []

        s = select([t_point]).where(and_(*filters))

        heap = list()
        for row in engine.execute(s):
            dist = great_circle((lat, lng), (row.lat, row.lng))
            if dist <= radius:
                heap.append((dist, row.data))

        # Use heap sort to find top-K nearest
        n_nearest = heapq.nsmallest(n, heap, key=lambda x: x[0])
        return n_nearest


#--- Unittest ---
def test_geo_search():
    point_data = [
        (1, 38.42522236, -93.29622264),
        (2, 40.22483194, -92.37676291),
        (3, 39.71873925, -91.76397842),
    ]

    search_engine = GeoSearchEngine()
    search_engine.train(point_data,
                        key_id=lambda x: x[0],
                        key_lat=lambda x: x[1],
                        key_lng=lambda x: x[2],
                        )

    def test_find_n_nearest():
        n_nearest = search_engine.find_n_nearest(39.0, -92.0, n=1)
        assert len(n_nearest) == 1

        n_nearest = search_engine.find_n_nearest(39.0, -92.0, radius=60.0)
        for dist, point in n_nearest:
            assert dist <= 60.0

        n_nearest = search_engine.find_n_nearest(39.0, -92.0)
        assert len(n_nearest) == len(point_data)
        last_dist = -1.0
        for dist, point in n_nearest:
            assert last_dist <= dist
            last_dist = dist

    test_find_n_nearest()


if __name__ == "__main__":
    test_geo_search()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module is to make regular expression easier to use.
With some built-in compiled pattern, we can use human language-like syntax to 
generate re pattern.
"""

import re


def extract_by_prefix_surfix(text, prefix, surfix, minlen=None, maxlen=None,
                             include=False):
    """Extract the text in between a prefix and surfix. It use non-greedy match.

    :param text: text body
    :type text: str

    :param prefix: the prefix
    :type prefix: str

    :param surfix: the surfix
    :type surfix: str

    :param minlen: the min matched string length
    :type minlen: int

    :param maxlen: the max matched string length
    :type maxlen: int

    :param include: whether if include prefix and surfix
    :type include: bool
    """
    if minlen is None:
        minlen = 0
    if maxlen is None:
        maxlen = 2 ** 30
    pattern = r"""(?<=%s)[\s\S]{%s,%s}?(?=%s)""" % (
        prefix, minlen, maxlen, surfix)
    if include:
        return [prefix + s + surfix for s in re.findall(pattern, text)]
    else:
        return re.findall(pattern, text)


def extract_number(text):
    """Extract digit character from text.
    """
    result = list()
    chunk = list()
    valid_char = set(".1234567890")
    for char in text:
        if char in valid_char:
            chunk.append(char)
        else:
            result.append("".join(chunk))
            chunk = list()
    result.append("".join(chunk))

    result_new = list()
    for number in result:
        if "." in number:
            try:
                result_new.append(float(number))
            except:
                pass
        else:
            try:
                result_new.append(int(number))
            except:
                pass

    return result_new


_regex_extract_email = re.compile(
    r"""([a-z0-9!#$%&'*+\/=?^_`{|}~-]+(?:\.[a-z0-9!#$%&'*+\/=?^_`{|}~-]+)*(@|\sat\s)(?:[a-z0-9](?:[a-z0-9-]*[a-z0-9])?(\.|\sdot\s))+[a-z0-9](?:[a-z0-9-]*[a-z0-9])?)""")
_regex_validate_email = re.compile(
    r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)")


def extract_email(text):
    """Extract email from text.
    """
    result = list()
    for tp in re.findall(_regex_extract_email, text.lower()):
        for email in tp:
            if re.match(_regex_validate_email, email):
                result.append(email)
    return result


_regex_url = re.compile(
    r"""(?i)\b((?:[a-z][\w-]+:(?:/{1,3}|[a-z0-9%])|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'".,<>?«»“”‘’]))""")

_regex_web_url = re.compile(
    r"""(?i)\b((?:https?:(?:/{1,3}|[a-z0-9%])|[a-z0-9.\-]+[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)/)(?:[^\s()<>{}\[\]]+|\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\))+(?:\([^\s()]*?\([^\s()]+\)[^\s()]*?\)|\([^\s]+?\)|[^\s`!()\[\]{};:'".,<>?«»“”‘’])|(?:(?<!@)[a-z0-9]+(?:[.\-][a-z0-9]+)*[.](?:com|net|org|edu|gov|mil|aero|asia|biz|cat|coop|info|int|jobs|mobi|museum|name|post|pro|tel|travel|xxx|ac|ad|ae|af|ag|ai|al|am|an|ao|aq|ar|as|at|au|aw|ax|az|ba|bb|bd|be|bf|bg|bh|bi|bj|bm|bn|bo|br|bs|bt|bv|bw|by|bz|ca|cc|cd|cf|cg|ch|ci|ck|cl|cm|cn|co|cr|cs|cu|cv|cx|cy|cz|dd|de|dj|dk|dm|do|dz|ec|ee|eg|eh|er|es|et|eu|fi|fj|fk|fm|fo|fr|ga|gb|gd|ge|gf|gg|gh|gi|gl|gm|gn|gp|gq|gr|gs|gt|gu|gw|gy|hk|hm|hn|hr|ht|hu|id|ie|il|im|in|io|iq|ir|is|it|je|jm|jo|jp|ke|kg|kh|ki|km|kn|kp|kr|kw|ky|kz|la|lb|lc|li|lk|lr|ls|lt|lu|lv|ly|ma|mc|md|me|mg|mh|mk|ml|mm|mn|mo|mp|mq|mr|ms|mt|mu|mv|mw|mx|my|mz|na|nc|ne|nf|ng|ni|nl|no|np|nr|nu|nz|om|pa|pe|pf|pg|ph|pk|pl|pm|pn|pr|ps|pt|pw|py|qa|re|ro|rs|ru|rw|sa|sb|sc|sd|se|sg|sh|si|sj|Ja|sk|sl|sm|sn|so|sr|ss|st|su|sv|sx|sy|sz|tc|td|tf|tg|th|tj|tk|tl|tm|tn|to|tp|tr|tt|tv|tw|tz|ua|ug|uk|us|uy|uz|va|vc|ve|vg|vi|vn|vu|wf|ws|ye|yt|yu|za|zm|zw)\b/?(?!@)))""")
"""ref: https://github.com/rcompton/ryancompton.net/blob/master/assets/praw_drugs/urlmarker.py
"""


def extract_web_url(text):
    """Extract url from text.
    """
    return re.findall(_regex_web_url, text)


_regex_date_iso = "\d{4}-\d{1,2}-\d{1,2}"


def extract_date_iso(text):
    """Extract iso date format (%yyyy-%mm-%dd) from text.
    """
    return re.findall(_regex_date_iso, text)


_regex_date_us = "\d{1,2}/\d{1,2}/\d{2,4}"


def extract_date_us(text):
    """Extract US format (%m/%d/%y) from text
    """
    return re.findall(_regex_date_us, text)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from fuzzywuzzy import process


def choose_best(text, choice, criterion=None):
    if criterion is None:
        return choose_best(text, choice, criterion=0)
    else:
        res, confidence_level = process.extractOne(text, choice)
        if confidence_level >= criterion:
            return res
        else:
            return None


if __name__ == "__main__":
    choice = ["Atlanta Falcons", "New Cow Jets",
              "Tom boy", "New York Giants", "Dallas Cowboys"]
    text = "cowboy"
    res = choose_best(text, choice)
    print(res)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

**中文文档**

codestats是一个用于统计项目代码的小工具。

1. 首先需要从多个项目目录中, 筛选出相关的文件, 例如 ".py" 的Python文件。
2. 其次需要对


- .py: Python Script
- .cs: CSharp Script
- .html: Html
- .css: CSS
- .js: JavaScript
- text: UTF-8 Pure Text File
"""

from __future__ import print_function
import re
import os
from os.path import join, splitext, getsize
from dataIO import js
from filetool.files import WinFile, FileCollection


class Stats(object):

    def __init__(self, **kwargs):
        for attr, value in kwargs.items():
            object.__setattr__(self, attr, value)


class Filter(object):

    @staticmethod
    def filter(abspath):
        raise NotImplementedError


class FilterPython(Filter):
    __label__ = "Python"

    @staticmethod
    def filter(abspath):
        if splitext(abspath)[1].lower() == ".py":
            return True
        else:
            return False


def find_files(dir_path, filters):
    """

    :param dir_path:
    :param filters:
    """
    result = {f.__label__: [] for f in filters}

    for current_dir, dirname_list, basename_list in os.walk(dir_path):
        for basename in basename_list:
            abspath = join(dirname, basename)
            for f in filters:
                if f.filter(abspath):
                    result[f.__label__].append(abspath)

    return result


def arg_preprocess(arg):
    if isinstance(arg, list):
        return arg
    else:
        return [arg, ]


class CodeStats(object):

    """A simple Python project code, comment, docstr line counter.

    Code stats analyzer should be initiated with the project workspace path and
    the programming language. 

    - code: number of code lines, includes docstr.
    - comment: number of comment lines.
    - docstr: number of doc string lines.
    - purecode: code lines not include docstr.

    Usage example::

        >>> from angora.gadget import CodeStats
        >>> analyzer = CodeStats(workspace=r"C:\Python33\lib\site-packages\requests")
        >>> analyzer.run()
        Code statistic result for 'C:\Python33\lib\site-packages\requests'
          79 'Python' files, 80 other files.
            code line: 12362
            comment line: 2025
            docstr line: 1545
            purecode line: 10817

    **中文文档**

    :class:`CodeStats` 是一个用来统计Python项目中所有文件的代码, 注释和
    文档字符串行数的小工具。下面是对这些名词的说明:

    - code (代码): 代码行数, 包括文档字符串。
    - comment (注释): 注释行数。
    - docstr (文档字符串): 文档字符串行数。
    - purecode (纯代码): 纯代码不包括文档字符串。

    用例::

        >>> from angora.gadget import CodeStats
        >>> analyzer = CodeStats(workspace=r"C:\Python33\lib\site-packages\requests")
        >>> analyzer.run()
        Code statistic result for 'C:\Python33\lib\site-packages\requests'
          79 'Python' files, 80 other files.
            code line: 12362
            comment line: 2025
            docstr line: 1545
            purecode line: 10817        
    """

    def __init__(self, workspace, ignore=list()):
        if not os.path.exists(workspace):
            raise FileNotFoundError("%r doesn't exists!" % workspace)
        self.workspace = os.path.abspath(workspace)
        self.ignore = ignore
        self.language = "Python"
        self.analyzer = self.analyzePython
        self.filter = self.filterPython

    def run(self):
        """Run analysis.

        The basic idea is to recursively find all script files in specific 
        programming language, and analyze each file then sum it up.
        """
        n_target_file, n_other_file = 0, 0

        code, comment, docstr, purecode = 0, 0, 0, 0

        fc = FileCollection.from_path_except(self.workspace, self.ignore)

        fc_yes, fc_no = fc.select(self.filter, keepboth=True)
        n_other_file += len(fc_no)

        for abspath in fc_yes:
            try:
                with open(abspath, "rb") as f:
                    code_text = f.read().decode("utf-8")
                    code_, comment_, docstr_, purecode_ = self.analyzer(
                        code_text)
                    code += code_
                    comment += comment_
                    docstr += docstr_
                    purecode += purecode_
                    n_target_file += 1
            except Exception as e:
                n_other_file += 1

        lines = list()
        lines.append("Code statistic result for '%s'" % self.workspace)
        lines.append("  %r %r files, %r other files." %
                     (n_target_file, self.language, n_other_file))
        lines.append("    code line: %s" % code)
        lines.append("    comment line: %s" % comment)
        lines.append("    docstr line: %s" % docstr)
        lines.append("    purecode line: %s" % purecode)
        message = "\n".join(lines)

        print(message)
        return message

    @staticmethod
    def filterPython(winfile):
        if winfile.ext == ".py":
            return True
        else:
            return False

    @staticmethod
    def analyzePython(code_text):
        """Count how many line of code, comment, dosstr, purecode in one 
        Python script file.
        """
        code, comment, docstr = 0, 0, 0

        p1 = r"""(?<=%s)[\s\S]*?(?=%s)""" % ('"""', '"""')
        p2 = r"""(?<=%s)[\s\S]*?(?=%s)""" % ("'''", "'''")

        # count docstr
        for pattern in [p1, p2]:
            for res in re.findall(pattern, code_text)[::2]:
                lines = [i.strip() for i in res.split("\n") if i.strip()]
                docstr += len(lines)

        # count comment line and code
        lines = [i.strip() for i in code_text.split("\n") if i.strip()]
        for line in lines:
            if line.startswith("#"):
                comment += 1
            else:
                code += 1
        purecode = code - docstr  # pure code = code - docstr
        return code, comment, docstr, purecode

    def forPython(self):
        self.filter = self.filterPython
        self.analyzer = self.analyzePython
        self.language = "Python"

    @staticmethod
    def filterText(winfile):
        return True

    @staticmethod
    def analyzeText(code_text):
        code, commend, docstr = 0, 0, 0
        code = code_text.count("\n") + 1
        comment = 0
        docstr = 0
        purecode = code - docstr  # pure code = code - docstr
        return code, comment, docstr, purecode

    def forText(self):
        self.filter = self.filterText
        self.analyzer = self.analyzeText
        self.language = "Text"


# #--- Unittest ---
if __name__ == "__main__":
    import os

    workspace = os.path.dirname(__file__)

    analyzer = CodeStats(workspace)

    analyzer.forPython()
    analyzer.run()

    analyzer.forText()
    analyzer.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
import time


class Timer(object):
    """A context manager style timer to measure execution time.

    **中文文档**

    一个简单的计时器, 提供了Context Manager语法。

    Example1::

        # Context Manager
        n = 1000 * 1000
        l = list(range(n))

        title = "C++ Style for loop"
        # Context Manager Syntax
        with Timer(title, display=True) as timer:
            for index in range(n):
                l[index]

    请注意, 该Timer的实现方式仅仅是用两个time.clock()相减, 效果并不会太好。
    """

    def __init__(self, display=True, title=None):
        self._start = None
        self._end = None
        self._elapsed = None
        self._display = display
        self._template = "elapsed {elapsed:.6f} second."
        self._title = title
        if title:
            self._template = title + ": " + self._template

    def start(self):
        self._start = time.clock()

    def end(self):
        self._end = time.clock()
        self._elapsed = self._end - self._start
        if self._display:
            self.display()

    def display(self):
        print(self._template.format(elapsed=self._elapsed))

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc_info):
        self.end()

    @property
    def elapsed(self):
        return self._elapsed


class EasyTimer(object):
    """A Timer can measure execution time of multiple block of time, respectively.
    """

    def __init__(self):
        self.st = None
        self.ed = None
        self.elapsed = None
        self.records = list()

    @property
    def total_elapsed(self):
        return sum(self.records)

    # single time, multiple time measurement
    def start(self):
        """Start measuring.
        """
        self.st = time.clock()

    def stop(self):
        """Save last elapse time to self.records.
        """
        self.ed = time.clock()
        self.elapsed = self.ed - self.st
        self.records.append(self.elapsed)

    def timeup(self):
        """Print the last measurement elapse time, and return it.
        """
        self.stop()
        self.display()

    def click(self):
        """Record the last elapse time and start the next measurement.
        """
        self.stop()
        self.start()

    def display(self):
        """Print the last elapse time.
        """
        print("Elapsed %.6f seconds" % self.elapsed)

    def reset(self):
        """Reset the timer.
        """
        self.elapsed = 0.0
        self.records = list()


def timeit(func, n, display=True, title=None, *args, **kwargs):
    """Measure a function's execution time, repeat n times and take the average.

    **中文文档**

    一个测试函数运行时间的小函数, 用于重复运行某函数若干次, 返回运行时间的
    平均值。

    :param func: 待测试的函数
    :param n: 重复的次数
    :param display: 是否打印运行时间
    :param template: 打印信息的模板
    :param args, kwargs: 待测函数的参数

    :returns elapsed: 运行所使用的时间
    """
    template = "elapsed {elapsed:.6f} second, repeat {n} times."
    if title:
        template = title + ": " + template

    st = time.clock()
    for i in range(n):
        func(*args, **kwargs)
    elapsed = (time.clock() - st) / n
    if display:
        print(template.format(elapsed=elapsed, n=n))
    return elapsed



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
collection of usefule decorator.
"""

from __future__ import print_function
import time


def _text_of_func_args_and_kwargs(func, args, kwargs):
    """

    **中文文档**

    返回一个函数的调用以及其参数的文本形式。
    """
    text_args = ", ".join(["%r" % arg for arg in args])
    text_kwargs = ", ".join(["%s=%r" % (key, value)
                             for key, value in kwargs.items()])
    if text_args and text_kwargs:
        return "%s(%s, %s)" % (func.__name__, text_args, text_kwargs)
    elif (not text_args) and (not text_kwargs):
        return "%s()" % func.__name__
    elif text_args:
        return "%s(%s)" % (func.__name__, text_args)
    elif text_kwargs:
        return "%s(%s)" % (func.__name__, text_kwargs)
    else:
        raise Exception


def elapsed_printer(func):
    """

    **中文文档**

    此包装器可以打印函数的输入参数, 以及运行时间。
    """
    def _wrapper(*args, **kwargs):
        print(">>> %s # Running ..." %
              _text_of_func_args_and_kwargs(func, args, kwargs))
        st = time.clock()
        res = func(*args, **kwargs)
        elapsed = time.clock() - st
        print("    Complete! Elapsed %.6f seconds." % elapsed)
        return res

    return _wrapper


def run_if_is_main(__name__):
    """

    **中文文档**

    此装饰器能够让函数自动只在自己是以主脚本进行时才运行, 否则直接返回None。
    此装饰器在pytest下无效。
    """
    def _run_if_is_main(func):
        def _wrapper(*args, **kwargs):
            if __name__ == "__main__":
                return func(*args, **kwargs)
        return _wrapper

    return _run_if_is_main


@run_if_is_main(__name__)
def test_run_if_is_main():
    return "HelloWorld"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from pathlib_mate import Path


def filter_python_script(path):
    if path.ext == ".py":
        return True
    else:
        return False


def filter_pure_text(path):
    ext = [".txt", ".rst", ".md"]
    if path.ext in ext:
        return True
    else:
        return False


def count_lines(abspath):
    """Count how many lines in a pure text file.
    """
    with open(abspath, "rb") as f:
        i = 0
        for line in f:
            i += 1
            pass
        return i


def lines_stats(dir_path, file_filter):
    """Lines count of selected files under a directory.

    :return n_files: number of files
    :return n_lines: number of lines
    """
    n_files = 0
    n_lines = 0
    for p in Path(dir_path).select_file(file_filter):
        n_files += 1
        n_lines += count_lines(p.abspath)
    return n_files, n_lines



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""

**中文文档**

DictTree是类似于xml的一种文档树结构的实现。

- set attributes 操作为DictTree添加metadata, 例如 ``DictTree.name = "John"``
- set item 操作为DictTree添加子树, 例如 ``DictTree["Sam"] = DictTree(name="Sam")``
- for 循环操作为遍历所有的子树的key, 同理有 keys(), values(), items() 方法可以对
  key, value或者两者进行循环遍历。
- keys_at(), values_at(), items_at() 可以根据树的深度进行遍历。深度为0返回母树本身。
- length_at() 返回在某个深度上子树的数量。
- stats(), stats_at() 返回树的基本信息统计情况。

注意: 本实现性能很一般, 若你对性能有需求, 请使用xml文档树。本实现是面向对象的
实现方式, 根据同样的思路, 还有一个函数式的实现 :mod:`~sfm.dicttree`. 缺点是操作
比较不直观, 但是对于stats方法的性能有巨大的提升。这是因为在返回子树时候无需生成
子树对象。虽然生成子树对象实际上仅仅是将类的属性和字典进行绑定, 但是相对于在
循环中计数这一快速过程, 仍然是一笔不小的开销。
"""

import json
import pickle
import collections


class DictTree(object):
    """An pure python xml doc tree implementation in dictionary.

    **中文文档**

    internal data structure::

        {
            "__meta__": {key: value}, # parent tree attributes
            "child_key1": ... , # child tree's key, value pair.
            "child_key2": ... ,
            ...
        }

    对于根树而言, Key为 ``"root"``
    """
    __slots__ = ["__data__", ]

    def __init__(self, __data__=None, **kwargs):
        if __data__ is None:
            object.__setattr__(self, "__data__", {"__meta__": kwargs})
        else:
            object.__setattr__(self, "__data__", __data__)

    def __str__(self):
        try:
            return json.dumps(self.__data__, sort_keys=True, indent=4)
        except:
            return str(self.__data__)

    def dump(self, path):
        """dump DictTree data to json files.
        """
        try:
            with open(path, "wb") as f:
                f.write(self.__str__().encode("utf-8"))
        except:
            pass

        with open(path, "wb") as f:
            pickle.dump(self.__data__, f)

    @staticmethod
    def load(path):
        """load DictTree from json files.
        """
        try:
            with open(path, "rb") as f:
                return DictTree(__data__=json.loads(f.read().decode("utf-8")))
        except:
            pass

        with open(path, "rb") as f:
            return DictTree(__data__=pickle.load(f))

    def __getattribute__(self, attr):
        """
        """
        try:
            return object.__getattribute__(self, "__data__")["__meta__"][attr]
        except KeyError:
            return object.__getattribute__(self, attr)

    def __setattr__(self, attr, value):
        """
        """
        self.__data__["__meta__"][attr] = value

    def __setitem__(self, key, dicttree):
        if key == "__meta__":
            raise ValueError("'key' can't be '__meta__'!")

        if isinstance(dicttree, DictTree):
            self.__data__[key] = dicttree.__data__
        else:
            raise TypeError("attribute assignment only takes 'DictTree'.")

    def __getitem__(self, key):
        return DictTree(__data__=self.__data__[key])

    def __delitem__(self, key):
        if key == "__meta__":
            raise ValueError("'key' can't be '__meta__'!")

        del self.__data__[key]

    def __len__(self):
        """Return number of child trees.

        **中文文档**

        返回子树的数量。
        """
        return len(self.__data__) - 1

    def __iter__(self):
        for key in self.__data__:
            if key != "__meta__":
                yield key

    def keys(self):
        for key in self.__data__:
            if key != "__meta__":
                yield key

    def values(self):
        for key, value in self.__data__.items():
            if key != "__meta__":
                yield DictTree(__data__=value)

    def items(self):
        for key, value in self.__data__.items():
            if key != "__meta__":
                yield key, DictTree(__data__=value)

    def keys_at(self, depth, counter=1):
        if depth < 1:
            yield "__root__"
        else:
            if counter == depth:
                for key in self.keys():
                    yield key
            else:
                counter += 1
                for dicttree in self.values():
                    for key in dicttree.keys_at(depth, counter):
                        yield key

    def values_at(self, depth):
        if depth < 1:
            yield self
        else:
            for dicttree in self.values():
                for value in dicttree.values_at(depth - 1):
                    yield value

    def items_at(self, depth):
        if depth < 1:
            yield "__root__", self
        elif depth == 1:
            for key, value in self.items():
                yield key, value
        else:
            for dicttree in self.values():
                for key, value in dicttree.items_at(depth - 1):
                    yield key, value

    def length_at(self, depth):
        """Get the number of nodes on specific depth.
        """
        if depth == 0:
            return 1

        counter = 0
        for dicttree in self.values_at(depth - 1):
            counter += len(dicttree)
        return counter

    def stats(self, result=None, counter=0):
        """Display the node stats info on specific depth in this dict.

        ::

            [
                {"depth": 0, "leaf": M0, "root": N0},
                {"depth": 1, "leaf": M1, "root": N1},
                ...
                {"depth": k, "leaf": Mk, "root": Nk},
            ]
        """
        if result is None:
            result = dict()

        if counter == 0:
            if len(self):
                result[0] = {"depth": 0, "leaf": 0, "root": 1}
            else:
                result[0] = {"depth": 0, "leaf": 1, "root": 0}

        counter += 1
        if len(self):
            result.setdefault(
                counter, {"depth": counter, "leaf": 0, "root": 0})
            for dicttree in self.values():
                if len(dicttree):  # root
                    result[counter]["root"] += 1
                else:  # leaf
                    result[counter]["leaf"] += 1
                dicttree.stats(result, counter)

        return [
            collections.OrderedDict([
                ("depth", info["depth"]),
                ("leaf", info["leaf"]),
                ("root", info["root"]),
            ]) for info in sorted(result.values(), key=lambda x: x["depth"])
        ]

    def stats_at(self, depth):
        root, leaf = 0, 0
        for dicttree in self.values_at(depth):
            if len(dicttree):
                root += 1
            else:
                leaf += 1
        total = root + leaf
        print("On depth %s, having %s root nodes, %s leaf nodes. "
              "%s nodes in total." % (depth, root, leaf, total))


if __name__ == "__main__":
    def test_performance():
        path = "test.json"

        st = time.clock()
        d = DictTree(name=rand_str(8))
        for depth in range(6):
            for dicttree in d.values_at(depth):
                for _ in range(10):
                    dicttree[rand_str(8)] = DictTree(name=rand_str(8))
        print("creating elapse %.6f" % (time.clock() - st,))

        st = time.clock()
        d.dump(path)
        print("dumping elapse %.6f" % (time.clock() - st,))

        st = time.clock()
        d = DictTree.load(path)
        print("loading elapse %.6f" % (time.clock() - st,))

        st = time.clock()
        ppt(d.stats())
        print("analyze elapse %.6f" % (time.clock() - st,))

        st = time.clock()
        for depth in range(7):
            d.stats_at(depth)
        print("analyze elapse %.6f" % (time.clock() - st,))


# ....creating elapse 27.871160
# dumping elapse 23.268375
# loading elapse 3.634183
# [OrderedDict([('depth', 0), ('leaf', 0), ('root', 1)]),
#  OrderedDict([('depth', 1), ('leaf', 0), ('root', 10)]),
#  OrderedDict([('depth', 2), ('leaf', 0), ('root', 100)]),
#  OrderedDict([('depth', 3), ('leaf', 0), ('root', 1000)]),
#  OrderedDict([('depth', 4), ('leaf', 0), ('root', 10000)]),
#  OrderedDict([('depth', 5), ('leaf', 0), ('root', 100000)]),
#  OrderedDict([('depth', 6), ('leaf', 1000000), ('root', 0)])]
# analyze elapse 86.070060
# ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module is provide methods for searching item in a sorted list.

Highlight:

- :func:`find_last_true`: A magic pluggable method, read API reference for more
  info.
- :func:`find_nearest`: Find the nearest item of x from sorted array.

**中文文档**

下面是一些讨论:

原生的 ``bisect.bisect_left`` 和 ``bisect.bisect_right`` 方法返回的是被搜索元素的
index。而Python原生的list的getitem方法的效率并不够快, 所以当我们对一个有序大数组
重复搜索时, 我们最好使用一个: 元素有序, 访问复杂度为O(1)的数据结构。这个数据结构
是什么, 还留给读者思考。
"""

from __future__ import print_function
import bisect

########################################################################
# official recepi from doc.python.org                                  #
# https://docs.python.org/2/library/bisect.html#searching-sorted-lists #
########################################################################


def find_index(array, x):
    """Locate the leftmost value exactly equal to x.

    **中文文档**

    返回第一个值等于x的元素的索引。
    """
    i = bisect.bisect_left(array, x)
    if i != len(array) and array[i] == x:
        return i
    raise ValueError


def find_lt(array, x):
    """Find rightmost value less than x.

    Example::

        >>> find_lt([0, 1, 2, 3], 2.5)
        2

    **中文文档**

    寻找最大的小于x的数。
    """
    i = bisect.bisect_left(array, x)
    if i:
        return array[i - 1]
    raise ValueError


def find_le(array, x):
    """Find rightmost value less than or equal to x.

    Example::

        >>> find_le([0, 1, 2, 3], 2.0)
        2

    **中文文档**

    寻找最大的小于等于x的数。
    """
    i = bisect.bisect_right(array, x)
    if i:
        return array[i - 1]
    raise ValueError


def find_gt(array, x):
    """Find leftmost value greater than x.

    Example::

        >>> find_gt([0, 1, 2, 3], 0.5)
        1

    **中文文档**

    寻找最小的大于x的数。
    """
    i = bisect.bisect_right(array, x)
    if i != len(array):
        return array[i]
    raise ValueError


def find_ge(array, x):
    """Find leftmost item greater than or equal to x.

    Example::

        >>> find_ge([0, 1, 2, 3], 1.0)
        1

    **中文文档**

    寻找最小的大于等于x的数。
    """
    i = bisect.bisect_left(array, x)
    if i != len(array):
        return array[i]
    raise ValueError


def find_nearest(sorted_list, x):
    """Find the nearest item of x from sorted array.

    note: for finding the nearest item from a descending array, I recommend
    find_nearest(sorted_list[::-1], x). Because the built-in list[::-1] method
    is super fast.

    Usage::

        >>> find_nearest([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], 5.1)
        5

    **中文文档**

    在正序数组中, 返回最接近x的数。
    """
    if x <= sorted_list[0]:
        return sorted_list[0]
    elif x >= sorted_list[-1]:
        return sorted_list[-1]
    else:
        lower = find_le(sorted_list, x)
        upper = find_ge(sorted_list, x)
        if (x - lower) > (upper - x):
            return upper
        else:
            return lower


def find_last_true(sorted_list, true_criterion):
    """Suppose we have a list of item [item1, item2, ..., itemN].

    If we do a mapping::

        >>> def true_criterion(item):
        ...     return item <= 6
        >>> [true_criterion(item) for item in sorted_list]
        [True, True, ... True(last true), False, False, ... False]

    this function returns the index of last true item.

    we do can do the map for all item, and run a binary search to find the 
    index. But sometime the mapping function is expensive. This method avoid
    run mapping function for all items.

    Example::

        array     = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        index     = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        criterion = def true_criterion(x): return x <= 6
        boolean   = [1, 1, 1, 1, 1, 1, 1, 0, 0, 0]

    Solution::

        # first, we check index = int((0 + 9)/2.0) = 4, it's True. 
        # Then check array[4 + 1], it's still True. 
        # Then we jump to int((4 + 9)/2.0) = 6, it's True. 
        # Then check array[6 + 1], ite's False. So array[6] is the one we need.

        >>> find_last_true([0, 1, 2, 3, 4, 5, 6, 7, 8, 9], true_criterion)
        6

    **中文文档**

    功能: 假设有一组排序号了的元素, 从前往后假设前面的元素都满足某一条件, 而到了
    中间某处起就不再满足了。本函数返回满足这一条件的最后一个元素。这在当检验是否
    满足条件本身开销较大时, 能节约大量的计算时间。例如你要判定一系列网页中, 从
    page1 到 page999, 从第几页开始出现404错误。假设是第400个, 那么如果一个个地
    去试, 需要400次, 那如果从0 - 999之间去试, 只需要试验9次即可 (2 ** 9 = 512)

    算法: 

    我们检验最中间的元素, 如果为False, 那么则检验左边所有未检验过的元素的最中间
    的那个。如果为True, 那么检验右边所有未检验过的元素的最中间那个。重复这一过程
    直到被检验的元素为True, 而下一个元素为False, 说明找到了。


    例题::

        有序数组    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        序号        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
        条件        小于等于6
        真值表      [1, 1, 1, 1, 1, 1, 1, 0, 0, 0]

    解::

        第一次检查``index = int((0+9)/2.0) = 4``, 为True,
        检查array[4+1], 也是True。那么跳跃至``int((4+9)/2.0)=6``, 为True,。
        再检查array[6+1], 为False, 很显然, 我们找到了。
    """

    # exam first item, if not true, then impossible to find result
    if not true_criterion(sorted_list[0]):
        raise ValueError

    # exam last item, if true, it is the one.
    if true_criterion(sorted_list[-1]):
        return sorted_list[-1]

    lower, upper = 0, len(sorted_list) - 1

    index = int((lower + upper) / 2.0)
    while 1:
        if true_criterion(sorted_list[index]):
            if true_criterion(sorted_list[index + 1]):
                lower = index
                index = int((index + upper) / 2.0)
            else:
                return index
        else:
            upper = index
            index = int((lower + index) / 2.0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Similar to ``collections.namedtuple``, ``nameddict`` is a data container class.
Provides methods to iterate on attributes and values.

**中文文档**

和 ``collections.namedtuple`` 类似, ``nameddict`` 是一种数据容器类。提供了方便的方法
对属性, 值进行for循环, 以及和list, dict之间的IO交互。
"""

import json
import copy
from collections import OrderedDict
from functools import total_ordering


@total_ordering
class Base(object):

    """nameddict base class.
    """
    __attrs__ = None
    """该属性非常重要, 定义了哪些属性被真正视为 ``attributes``, 换言之, 就是在
    :meth:`~Base.keys()`, :meth:`~Base.values()`, :meth:`~Base.items()`,
    :meth:`~Base.to_list()`, :meth:`~Base.to_dict()`, :meth:`~Base.to_OrderedDict()`,
    :meth:`~Base.to_json()`, 方法中要被包括的属性。
    """

    __excludes__ = []
    """在此被定义的属性将不会出现在 :meth:`~Base.items()` 中
    """

    __reserved__ = set(["keys", "values", "items"])

    def __init__(self, **kwargs):
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    def __setattr__(self, attr, value):
        if attr in self.__reserved__:
            raise ValueError("%r is a reserved attribute name!" % attr)
        object.__setattr__(self, attr, value)

    def __repr__(self):
        kwargs = list()
        for attr, value in self.items():
            kwargs.append("%s=%r" % (attr, value))
        return "%s(%s)" % (self.__class__.__name__, ", ".join(kwargs))

    def __getitem__(self, key):
        """Access attribute.
        """
        return object.__getattribute__(self, key)

    @classmethod
    def _make(cls, d):
        """Make an instance.
        """
        return cls(**d)

    def items(self):
        """items按照属性的既定顺序返回attr, value对。当 ``__attrs__`` 未指明时,
        则按照字母顺序返回。若 ``__attrs__`` 已定义时, 按照其中的顺序返回。

        当有 ``@property`` 装饰器所装饰的属性时, 若没有在 ``__attrs__`` 中定义,
        则items中不会包含它。
        """
        items = list()

        if self.__attrs__ is None:
            for key, value in self.__dict__.items():
                if key not in self.__excludes__:
                    items.append((key, value))
            items = list(sorted(items, key=lambda x: x[0]))
            return items
        try:
            for attr in self.__attrs__:
                if attr not in self.__excludes__:
                    try:
                        items.append(
                            (attr, copy.deepcopy(getattr(self, attr))))
                    except AttributeError:
                        items.append(
                            (attr, copy.deepcopy(self.__dict__.get(attr))))
            return items
        except:
            raise AttributeError()

    def keys(self):
        """Iterate attributes name.
        """
        return [key for key, value in self.items()]

    def values(self):
        """Iterate attributes value.
        """
        return [value for key, value in self.items()]

    def __iter__(self):
        """Iterate attributes.
        """
        if self.__attrs__ is None:
            return iter(self.keys())
        try:
            return iter(self.__attrs__)
        except:
            raise AttributeError()

    def to_list(self):
        """Export data to list. Will create a new copy for mutable attribute.
        """
        return self.keys()

    def to_dict(self):
        """Export data to dict. Will create a new copy for mutable attribute.
        """
        return dict(self.items())

    def to_OrderedDict(self):
        """Export data to OrderedDict. Will create a new copy for mutable 
        attribute.
        """
        return OrderedDict(self.items())

    def to_json(self):
        """Export data to json. If it is json serilizable.
        """
        return json.dumps(self.to_dict())

    def __eq__(self, other):
        """Equal to.
        """
        return self.items() == other.items()

    def __lt__(self, other):
        """Less than.
        """
        for (_, value1), (_, value2) in zip(self.items(), other.items()):
            if value1 >= value2:
                return False
        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-


"""
function logic flow constructor.
"""


def try_ntime(max_try, func, *args, **kwargs):
    """Try execute a function n times, until no exception raised or tried 
    ``max_try`` times.

    **中文文档**

    反复尝试执行一个函数若干次。直到成功为止或是重复尝试 ``max_try`` 次。期间
    只要有一次成功, 就正常返回。如果一次都没有成功, 则行为跟最后一次执行了
    ``func(*args, **kwargs)`` 一样。
    """
    for i in range(max_try):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            current_exception = e
    raise current_exception



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from .pkg import sqlitedict
    from .scheduler import Task, BaseDBTableBackedScheduler
except:  # pragma: no cover
    from pytq.pkg import sqlitedict
    from pytq.scheduler import Task, BaseDBTableBackedScheduler


class SqliteDictScheduler(BaseDBTableBackedScheduler):
    user_db_path = None

    def __init__(self, logger=None, user_db_path=None):
        super(SqliteDictScheduler, self).__init__(logger=logger)

        # link encode method
        try:
            self.user_encode(None)
            encode = self.encode
        except NotImplementedError:
            encode = sqlitedict.encode
        except:
            encode = self.encode

        # link decode method
        try:
            self.user_decode(None)
            decode = self.decode
        except NotImplementedError:
            decode = sqlitedict.decode
        except:
            decode = self.decode

        # initiate back end database
        self._dct = sqlitedict.SqliteDict(
            self.user_db_path, autocommit=True,
            encode=encode,
            decode=decode,
        )

    @property
    def user_db_path(self):
        """
        Back-end sqlite database file path.
        """
        raise NotImplementedError

    def user_encode(self, obj):
        """
        (Optional) User defined serializer for output_data.

        :returns: bytes or string.

        **中文文档**

        用于对处理结果序列化的函数。默认使用pickle。
        """
        raise NotImplementedError

    def user_decode(self, bytes_or_str):
        """
        (Optional) User defined deserializer for output_data.

        :returns: python object.

        **中文文档**

        用于对处理结果反序列化的函数。默认使用pickle。
        """
        raise NotImplementedError

    def _default_is_duplicate(self, task):
        return task.id in self._dct

    def _get_finished_id_set(self):
        return set(self._dct.keys())

    def _default_post_process(self, task):
        self._dct[task.id] = task.output_data

    def __len__(self):
        return len(self._dct)

    def __iter__(self):
        return iter(self._dct)

    def clear_all(self):
        self._dct.clear()

    def get(self, id):
        return self._dct.get(id)
